"""
This module contains base classes (some of which are abstract) common to both **Source** and
**Ground Motion Model (GMM)** logic trees.
"""

import copy
import json
from abc import ABC
from dataclasses import asdict, dataclass, field, fields
from functools import reduce
from itertools import product
from operator import mul
from pathlib import Path
from typing import Any, Dict, Generator, Generic, Iterator, List, Optional, Type, TypeVar, Union

import dacite

import nzshm_model.logic_tree.helpers as helpers
from nzshm_model.psha_adapter import PshaAdapterInterface

from .branch import Branch, BranchType, CompositeBranch
from .correlation import LogicTreeCorrelations

# TODO:
# - move values to the base class?
# - move logic for creating light branchSet and LogicTree when creating FilteredBranchs to
#    the initializer of FilteredBranch?
# - FilteredBranch doesn't need to be a data class as it should not be serialized and doesn't contain many arguments
# - should we use FilteredBranch for correlation so the branches can be traced back to the BranchSet?

# https://github.com/python/mypy/issues/8495
# This can be done more simply with typing.Self in python3.11+
LogicTreeType = TypeVar("LogicTreeType", bound="LogicTree")
BranchSetType = TypeVar("BranchSetType", bound="BranchSet")
FilteredBranchType = TypeVar("FilteredBranchType", bound="FilteredBranch")


@dataclass
class BranchSet(Generic[BranchType]):
    """
    A group of branches that comprise their own sub-logic tree. Also known as a fault system logic
    tree (for source logic trees).

    Arguments:
        short_name:
        long_name:
        branches: the branches that make up the branch set
    """

    short_name: str = ''
    long_name: str = ''
    branches: List[BranchType] = field(default_factory=list)

    def __post_init__(self):
        helpers._validate_branchset_weights(self)

    def __str__(self) -> str:
        string = f'short_name: {self.short_name}, long_name: {self.long_name}\n'
        string += '======BRANCHES======\n'
        return string + '\n'.join([str(branch) for branch in self])

    def __iter__(self: BranchSetType) -> BranchSetType:
        self.__counter = 0
        return self

    def __next__(self) -> BranchType:
        if self.__counter >= len(self.branches):
            raise StopIteration
        else:
            self.__counter += 1
            return self.branches[self.__counter - 1]


@dataclass
class LogicTree(ABC, Generic[FilteredBranchType]):
    """
    Logic tree baseclass. Contains information about branch sets and correlations between branches of the branch sets.

    The correlations are enforced when forming composite branches (combinations of branches from different branch sets).

    Arguments:
        title: title of the logic tree
        version: version string
        branch_sets: list of branch sets that make up the logic tree
        correlations: any correlations between branches of the branch_sets

    """

    title: str = ''
    version: str = ''
    branch_sets: List[Any] = field(default_factory=list)
    correlations: LogicTreeCorrelations = field(default_factory=LogicTreeCorrelations)

    def __post_init__(self) -> None:
        # TODO: branch set short_names should be unique (see from_branches())
        helpers._validate_correlation_weights(self)

    def __setattr__(self, __name: str, __value: Any) -> None:
        super().__setattr__(__name, __value)
        if __name == "correlations":
            helpers._validate_correlation_weights(self)

    def __str__(self) -> str:
        string = f'LogicTree type: {type(self)}\n'
        string += f'title: {self.title}, version:{self.version}\n'
        string += '==========BRANCH SETS==========\n\n'
        string += '\n\n'.join([str(bs) for bs in self.branch_sets])
        return string

    def _composite_branches(self) -> Generator[CompositeBranch, None, None]:
        """
        Yields all composite (combined) branches of the branch_sets without applying correlations.

        Returns:
            composite_branches: the CompositeBranches of the combined logic tree BranchSets
        """
        for branches in product(*[branch_set.branches for branch_set in self.branch_sets]):
            yield CompositeBranch(branches=branches)

    @property
    def composite_branches(self) -> Generator[CompositeBranch, None, None]:
        """
        Yields all composite (combined) branches of the branch_sets enforcing correlations.

        Returns:
            composite_branches: the CompositeBranches of the combined logic tree BranchSets
        """
        for composite_branch in self._composite_branches():
            # if the comp_branch contains a branch listed as the 0th element of the correlations, only
            # yeild if the other branches are present
            # weight is automatically calculated by CompositeBranch if not explicitly assigned
            # (as we would with correlations)
            correlation_match = [branch_pri in composite_branch for branch_pri in self.correlations.primary_branches()]
            if any(correlation_match):
                # index of the correlation that matches a branch in _composite_branches()
                i_cor = correlation_match.index(True)
                if not all(br in composite_branch for br in self.correlations[i_cor].all_branches):
                    continue
                weights = [self.correlations[i_cor].weight] + [
                    branch.weight for branch in composite_branch if branch not in self.correlations[i_cor].all_branches
                ]
                composite_branch.weight = reduce(mul, weights, 1.0)
            yield composite_branch

    @classmethod
    def from_json(cls: Type[LogicTreeType], json_path: Union[Path, str]) -> LogicTreeType:
        """
        Create LogicTree object from json file

        See docs/api/logic_tree/source_logic_tree_config_format.md and
        api/logic_tree/gmcm_logic_tree_config_format.md

        Parameters:
            json_path: path to json file

        Returns:
            logic_tree
        """
        with Path(json_path).open() as jsonfile:
            data = json.load(jsonfile)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls: Type[LogicTreeType], data: Dict) -> LogicTreeType:
        """
        Create LogicTree object from dict.

        See docs/api/logic_tree/source_logic_tree_config_format.md and
        api/logic_tree/gmcm_logic_tree_config_format.md

        Parameters:
            data: dict representation of LogicTree object

        Returns:
            logic_tree
        """
        if not data.get('correlations'):
            logic_tree = cls._from_dict(data)
            # do not need to validate names as that is only necessary if there are correlations
            return logic_tree

        correlations = data.pop('correlations')
        helpers._validate_correlations_format(correlations)
        logic_tree = cls._from_dict(data)
        helpers._validate_names(logic_tree)
        logic_tree = helpers._add_corellations(logic_tree, correlations)

        return logic_tree

    @classmethod
    def _from_dict(cls: Type[LogicTreeType], data: Dict) -> LogicTreeType:
        """
        Create LogicTree object from dict. Input dict must be a direct asdict() serialisation of the LogicTree object

        Parameters:
            data: dict representation of LogicTree object

        Returns:
            logic_tree
        """

        config = dacite.Config(strict=True, cast=[tuple])
        return dacite.from_dict(data_class=cls, data=data, config=config)

    def _to_dict(self) -> Dict[str, Any]:
        """
        Create dict representation of logic tree. This creates an exact dict of the class and is not
        used for serialisation.

        Returns:
            dict:
        """
        return asdict(self)

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary representation of logic tree used for serialisation.

        Returns:
            logic_tree_dict: dictionary representaion of logic tree
        """

        # do not need to validate names as that is only necessary if there are correlations
        data = self._to_dict()
        if not self.correlations:
            del data["correlations"]
            return data

        helpers._validate_names(self)
        data["correlations"] = helpers._serialise_correlations(self)
        return data

    def to_json(self, file_path: Union[Path, str]) -> None:
        """Serialze logic tree as json file.

        Parameters:
            file_path: path to json file to be written
        """
        file_path = Path(file_path)
        with file_path.open('w') as jsonfile:
            json.dump(self.to_dict(), jsonfile, indent=2)

    def __all_branches__(self) -> Generator[FilteredBranchType, None, None]:
        """
        Yield all branches from all BranchSets, each with a shallow copy of its LogicTree and BranchSet parents
        for use in filtering.

        NB this class is never used for serialising models.
        """

        def get_fields(obj):
            return {
                field.name: copy.deepcopy(getattr(obj, field.name))
                for field in fields(obj)
                if field.name not in ('branches', 'branch_sets')
            }

        lt_fields = get_fields(self)
        for branch_set in self.branch_sets:
            bs_fields = get_fields(branch_set)
            for branch in branch_set.branches:
                bs_lite = type(branch_set)(**bs_fields, branches=[type(branch)()])
                lt_lite = type(self)(**lt_fields)
                # b_fields = get_fields(branch)
                yield branch.filtered_branch(
                    branch_set=bs_lite,
                    logic_tree=lt_lite,
                )

    @classmethod
    def from_branches(cls, branches: Iterator[FilteredBranchType]) -> 'LogicTree':
        """
        Build a complete LogicTree from a iterable of branches.

        We expect that all the branches have come from a single LogicTree.

        Parameters:
            branches: the branches used to build the LogicTree
        """

        def match_branch_set(slt: LogicTree, fb):
            for branch_set in slt.branch_sets:
                if fb.branch_set.short_name == branch_set.short_name:
                    return branch_set

        version = None
        for fb in branches:
            # ensure an slt
            if version is None:
                logic_tree = cls(version=fb.logic_tree.version, title=fb.logic_tree.title)
                version = fb.logic_tree.version
            else:
                assert version == fb.logic_tree.version

            # ensure an branch_set
            bs = match_branch_set(logic_tree, fb)
            if not bs:
                bs = type(fb.branch_set)(short_name=fb.branch_set.short_name, long_name=fb.branch_set.long_name)
                logic_tree.branch_sets.append(bs)
            bs.branches.append(fb.to_branch())
        return logic_tree

    def __iter__(self):
        self.__current_branch = 0
        self.__branch_list = list(self.__all_branches__())
        return self

    def __next__(self) -> FilteredBranchType:
        if self.__current_branch >= len(self.__branch_list):
            raise StopIteration
        else:
            self.__current_branch += 1
            return self.__branch_list[self.__current_branch - 1]

    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs: Optional[Dict]) -> "PshaAdapterInterface":
        """get a PSHA adapter for this instance.

        Arguments:
            provider: the adapter class
            **kwargs: additional arguments required by the provider class

        Returns:
            a PSHA Adapter instance
        """
        return provider(target=self)


@dataclass
class FilteredBranch(Branch):
    """
    A branch type that points back to it's logic tree and branch set. Should never be serialized, only
    used for filtering
    """

    logic_tree: LogicTree = field(default_factory=LogicTree)
    branch_set: BranchSet = field(default_factory=BranchSet)

    def to_branch(self) -> Branch:
        """
        Produce the Branch object (does not point back to it's logic tree and branch set)

        Returns:
            branch: the Branch object
        """
        branch_attributes = {k: v for k, v in self.__dict__.items() if k not in ('branch_set', 'logic_tree')}
        return type(self.branch_set.branches[0])(**branch_attributes)



"""
Defines source logic tree structures used in NSHM.
"""

import copy
import warnings
from dataclasses import dataclass, field
from typing import List, Tuple, Union

from nzshm_model.logic_tree.correlation import Correlation, LogicTreeCorrelations
from nzshm_model.logic_tree.logic_tree_base import Branch, BranchSet, FilteredBranch, LogicTree

from . import BranchAttributeValue
from .fault_system_branch_set import BranchSetSpec
from .version1 import SourceLogicTree as SourceLogicTreeV1


@dataclass
class InversionSource:
    """
    A hazard source built from an NSHM Grand Inversion experiment

    Contains the specific identifiers used in NSHM Toshi API.

    Attributes:
        nrml_id: toshi_id for a NRML source file (needed by openquake)
        rupture_rate_scaling: the scaling ratio
        inversion_id: toshi_id for the source inversion solution
        rupture_set_id: toshi_id for the source rupture set
        inversion_solution_type: type (if any) of the source inversion solution
        type: constant "inversion"
    """

    nrml_id: str
    rupture_rate_scaling: Union[float, None] = None  # TODO: needed at this level??
    inversion_id: Union[str, None] = ""
    rupture_set_id: Union[str, None] = ""
    inversion_solution_type: Union[str, None] = ""
    type: str = "inversion"


@dataclass
class DistributedSource:
    """
    A gridded hazard source built from a background (off-fault) seismic rate model

    Contains the specific identifiers used in NSHM Toshi API.

    Attributes:
        nrml_id: toshi_id for a NRML source file (needed by openquake)
        rupture_rate_scaling: the scaling ratio
        type: constant "distributed"
    """

    nrml_id: str
    rupture_rate_scaling: Union[float, None] = None  # TODO: needed at this level??
    type: str = "distributed"


@dataclass
class SourceBranch(Branch):
    """
    A source branch can contain multiple sources.

    Contains the specific identifiers used in NSHM Toshi API.

    Attributes:
        values: list of attribute values that define the branch.
        sources: list of branch sources.
        rupture_rate_scaling: the scaling ratio.
    """

    values: List[BranchAttributeValue] = field(default_factory=list)
    sources: List[Union[DistributedSource, InversionSource]] = field(default_factory=list)
    rupture_rate_scaling: float = 1.0
    tectonic_region_types: Tuple[str, ...] = field(default_factory=tuple)

    def __post_init__(self):
        if not isinstance(self.tectonic_region_types, tuple):
            raise TypeError("tectonic_region_types must be a tuple")

    def filtered_branch(self, logic_tree: 'LogicTree', branch_set: 'BranchSet') -> 'FilteredBranch':
        """get a filtered branch containing reference to parent instances.

        Arguments:
            logic_tree: the source_logic_tree containing the branch
            branch_set: the branch_set containing the branch

        Returns:
            a SourceFilteredBranch instance
        """
        return SourceFilteredBranch(logic_tree=logic_tree, branch_set=branch_set, **self.__dict__)

    @property
    def tag(self) -> str:
        """tag identifies a branch based on its values (BranchAttributeValue).

        Returns:
            a string representation of the the list of values

        """
        return str(self.values)

    @property
    def registry_identity(self):
        nrmls = sorted([s.nrml_id for s in self.sources])
        return "|".join(nrmls)


# TODO: protect from users changing tectonic_region_types
@dataclass
class SourceBranchSet(BranchSet[SourceBranch]):
    """A list of Source Branches.

    Attributes:
        branches: list of branches.
    """

    branches: List[SourceBranch] = field(default_factory=list)

    def __post_init__(self):
        trts = {frozenset(branch.tectonic_region_types) for branch in self.branches}
        if len(trts) > 1:
            raise ValueError("all tectonic_region_types in a branch set must be the same")

    @property
    def tectonic_region_types(self) -> Tuple[str, ...]:
        return self.branches[0].tectonic_region_types if self.branches else ()


@dataclass
class SourceLogicTreeSpec:
    """Is this used anymore?"""

    branch_sets: List[BranchSetSpec] = field(default_factory=list)

    @property
    def fault_systems(self):
        """
        API alias for branch_sets (deprecated)
        """
        warnings.warn("Please use branch_sets property instead", DeprecationWarning)
        return self.branch_sets


@dataclass
class SourceLogicTree(LogicTree['SourceFilteredBranch']):
    """A dataclass representing a source logic tree

    Attributes:
        branch_sets: list of branch sets in this tree
        logic_tree_version: constant = 2

    """

    branch_sets: List[SourceBranchSet] = field(default_factory=list)  # branch_sets for this logic tree
    logic_tree_version: Union[int, None] = 2

    def __post_init__(self) -> None:

        # check that sources are defined correctly
        self._check_sources()
        super().__post_init__()

    def _check_sources(self):
        for branch in self:
            if not branch.sources:
                raise ValueError("every branch must have at least one source")
            for source in branch.sources:
                if isinstance(source, DistributedSource) and source.type != "distributed":
                    raise ValueError("source type DistributedSource does not match type member")
                if isinstance(source, InversionSource) and source.type != "inversion":
                    raise ValueError("source type DistributedSource does not match type member")

    def derive_spec(self) -> SourceLogicTreeSpec:
        raise NotImplementedError()
        # slt_spec = SourceLogicTreeSpec()
        # for fslt in self.fault_systems:
        #     slt_spec.fault_systems.append(FaultSystemLogicTree.derive_spec(fslt))
        # return slt_spec

    @staticmethod
    def from_source_logic_tree(original_slt: "SourceLogicTreeV1") -> "SourceLogicTree":
        """
        Migrate from old version one of slt.

        Arguments:
            original_slt: a v1 SourceLogicTree instance.

        Returns:
            a new SourceLogicTree instance
        """

        def index_branch(slt_v1, fslt_name, values):
            fslt_names = [fslt.short_name for fslt in slt_v1.fault_system_lts]
            ind_fslt = fslt_names.index(fslt_name)
            fslt_values = [branch.values for branch in original_slt.fault_system_lts[ind_fslt].branches]
            ind_branch = fslt_values.index(values)
            return ind_fslt, ind_branch

        if not isinstance(original_slt, SourceLogicTreeV1):
            raise ValueError(f"supplied object of {type(original_slt)} is not supported.")
        slt = SourceLogicTree(version=original_slt.version, title=original_slt.title)
        for fslt in original_slt.fault_systems:
            new_fslt = SourceBranchSet(short_name=fslt.short_name, long_name=fslt.long_name)
            branch_id = 0
            for branch in fslt.branches:
                new_branch = SourceBranch(
                    values=copy.deepcopy(branch.values), weight=branch.weight, branch_id=str(branch_id)
                )
                branch_id += 1
                if branch.onfault_nrml_id:
                    new_branch.sources.append(
                        InversionSource(
                            nrml_id=branch.onfault_nrml_id,
                            inversion_id=branch.inversion_solution_id,
                            rupture_set_id=branch.rupture_set_id,
                        )
                    )
                if branch.distributed_nrml_id:
                    new_branch.sources.append(DistributedSource(nrml_id=branch.distributed_nrml_id))
                new_fslt.branches.append(new_branch)
            slt.branch_sets.append(new_fslt)

        if original_slt.correlations:
            correlations = []
            for orig_correlation in original_slt.correlations:
                ind_fslt0, ind_branch0 = index_branch(
                    original_slt, orig_correlation.primary_short_name, orig_correlation.primary_values
                )
                ind_fslt1, ind_branch1 = index_branch(
                    original_slt, orig_correlation.secondary_short_name, orig_correlation.secondary_values
                )
                correlations.append(
                    Correlation(
                        primary_branch=slt.branch_sets[ind_fslt0].branches[ind_branch0],
                        associated_branches=[slt.branch_sets[ind_fslt1].branches[ind_branch1]],
                    )
                )
            slt.correlations = LogicTreeCorrelations(correlations)

        return slt

    @property
    def fault_system_lts(self):
        """
        API alias for branch_sets (deprecated)
        """
        warnings.warn("Please use branch_sets property instead", DeprecationWarning)
        return self.branch_sets

    @property
    def fault_systems(self):
        """
        API alias for branch_sets (deprecated)
        """
        warnings.warn("Please use branch_sets property instead", DeprecationWarning)
        return self.branch_sets


@dataclass
class SourceFilteredBranch(FilteredBranch, SourceBranch):
    """A logic tree source branch with additional properties

    Used to represent a branch that has been pruned (filtered) from a
    complete Source Logic Tree (SLT). The additional properties make the prundd branch
    traceable to its original SLT.

    Attributes:
        logic_tree: the source_logic_tree containing the branch
        branch_set: the branch_set containing the branch

    """

    logic_tree: 'LogicTree' = field(default_factory=SourceLogicTree)
    branch_set: 'BranchSet' = field(default_factory=SourceBranchSet)

    @property
    def fslt(self) -> 'BranchSet':
        """
        API alias for branch_set (deprecated)
        """
        warnings.warn("Please use branch_set property instead", DeprecationWarning)
        return self.branch_set

    @property
    def slt(self) -> 'LogicTree':
        """
        API alias for slt (deprecated)
        """
        warnings.warn("Please use logic_tree property instead", DeprecationWarning)
        return self.logic_tree



"""
The source_logic_tree package provides dataclasses that define the
Source Logic Tree dataclasses.
"""

from .branch_attribute import BranchAttributeSpec, BranchAttributeValue
from .logic_tree import InversionSource, SourceBranch, SourceBranchSet, SourceLogicTree
from .version1.logic_tree import SourceLogicTree as SourceLogicTreeV1
from .version1.logic_tree import SourceLogicTreeCorrelation



import math
from typing import TYPE_CHECKING, List

from .correlation import Correlation, LogicTreeCorrelations

if TYPE_CHECKING:
    from .logic_tree_base import BranchSet, LogicTree, LogicTreeType


##############################
# VALIDATORS
##############################
def _validate_branchset_weights(branch_set: 'BranchSet') -> None:
    """
    verify that weighs sum to 1.0
    """
    weight = 0.0
    if not branch_set.branches:  # empty BranchSet
        return
    for b in branch_set.branches:
        weight += b.weight
    if not math.isclose(weight, 1.0):
        raise ValueError("weights of BranchSet must sum to 1.0")


def _validate_names(logic_tree: 'LogicTree') -> None:
    # do not allow duplicate branch_set.shortname:branch.name
    branch_names = [f"{branch.branch_set.short_name}:{branch.branch_id}" for branch in logic_tree]
    if len(set(branch_names)) != len(branch_names):
        raise ValueError("branch_set.short_name:branch.branch_id must be unique")


def _validate_correlation_weights(logic_tree: 'LogicTree') -> None:
    # check that the weights total 1.0
    weight_total = 0.0
    for branch in logic_tree.composite_branches:
        weight_total += branch.weight
    if not math.isclose(weight_total, 1.0):
        raise ValueError("the weights of the logic tree do not sum to 1.0 when correlations are applied")


def _validate_correlations_format(correlations: List[str]) -> None:
    # check that the correlations are lists of stings with ":"
    for correlation in correlations:
        for branch in correlation:
            if not isinstance(branch, str):
                raise ValueError("branch names in correlations must be strings")
            if ":" not in branch:
                raise ValueError("names in correlations must be 'branch_set.shortname:branch.name' format")


##############################
# SERIALIZE / DESERIALIZE
##############################
def _correlation_encoding(branch_set, branch):
    return f"{branch_set.short_name}:{branch.branch_id}"


def _add_corellations(logic_tree: 'LogicTreeType', correlations: List[str]) -> 'LogicTreeType':

    branches = [branch for branch in logic_tree]
    # branch_names = [f"{fbranch.branch_set.short_name}:{fbranch.name}" for fbranch in branches]
    branch_names = [_correlation_encoding(fbranch.branch_set, fbranch) for fbranch in branches]
    correlation_groups = []
    for correlation in correlations:
        primary_branch = branches[branch_names.index(correlation[0])].to_branch()
        assoc_branches = [branches[branch_names.index(b)].to_branch() for b in correlation[1:]]
        correlation_groups.append(
            Correlation(
                primary_branch=primary_branch,
                associated_branches=assoc_branches,
            )
        )
    logic_tree.correlations = LogicTreeCorrelations(correlation_groups)

    return logic_tree


def _serialise_correlations(logic_tree: 'LogicTree') -> List[List[str]]:
    def find_branch_set(logic_tree, branch):
        for fbranch in logic_tree:
            if fbranch.to_branch() == branch:
                return fbranch.branch_set

    correlations = []
    for cor in logic_tree.correlations.correlation_groups:
        correlation = []
        correlation.append(_correlation_encoding(find_branch_set(logic_tree, cor.primary_branch), cor.primary_branch))
        for branch in cor.associated_branches:
            correlation.append(_correlation_encoding(find_branch_set(logic_tree, branch), branch))
        correlations.append(correlation)
    return correlations



"""
Classes for enforcing correlations between branches in logic trees
"""

import collections
import collections.abc
from dataclasses import dataclass, field
from typing import Generator, List, Optional, Sequence, overload

from .branch import Branch


@dataclass
class Correlation:
    """
    A correlation between **BranchSet**s of a logic tree. Correlations are used when buiding combinations of branches
    from differnt **BranchSet**s.

    For example, if a logic tree contains branchets A and B each with branches A1, A2 and B1, B2, respectivly, then
    a correlation A1, B1 will only allow A1-B1 not A1-B2 as valid composite branches.

    The primary_brach will not appear in combination with any branches except those identified in associated_branches
    for the relevent **BranchSet**s

    Arguments:
        primary_branch: the branch that **MUST** be correlated with the associated brances.
        associated_branches: list of branches that the primary_branch must always be coupled with.
        weight: weight used for composite branch formed by correlation. Defaults to weight of primary_branch
    """

    primary_branch: Branch = field(default_factory=Branch)
    associated_branches: List[Branch] = field(default_factory=list)
    weight: Optional[float] = None

    def __post_init__(self):
        self.weight = self.primary_branch.weight if not self.weight else self.weight

    @property
    def all_branches(self) -> List[Branch]:
        """
        list of all branches in a correlation; both the primary branch and the assiciated branches

        Returns:
            list of branches in correlation
        """
        return [self.primary_branch] + self.associated_branches


@dataclass(frozen=True)
class LogicTreeCorrelations(collections.abc.Sequence):
    """
    All correlations for a logic tree.

    Arguments:
        correlation_groups: list of correlations to be applied to the logic tree branch sets.
    """

    correlation_groups: List[Correlation] = field(default_factory=list)

    def __post_init__(self) -> None:
        _validate_correlations(self)

    def primary_branches(self) -> Generator[Branch, None, None]:
        """
        Yield the primary branches of all correlation_groups

        Returns:
            branches
        """
        for cor in self.correlation_groups:
            yield cor.primary_branch

    @overload
    def __getitem__(self, i: int) -> Correlation: ...  # noqa: E704

    @overload
    def __getitem__(self, i: slice) -> Sequence[Correlation]: ...  # noqa: E704

    def __getitem__(self, i):
        if isinstance(i, slice):
            raise TypeError("LogicTreeCorrelations does not support slicing")
        return self.correlation_groups[i]

    def __len__(self) -> int:
        return len(self.correlation_groups)


def _validate_correlations(ltcs: LogicTreeCorrelations) -> None:
    """
    check that there are no repeats in the 0th element of each correlation
    """
    prim_branches = list(ltcs.primary_branches())
    if len([branch for branch in prim_branches if prim_branches.count(branch) > 1]) != 0:
        raise ValueError("there is a repeated branch in the 0th element of the correlations")



import dataclasses
import io
import json
import logging
import sys

import click

import nzshm_model
from nzshm_model import all_model_versions, branch_registry, get_model_version

log = logging.getLogger()
logging.basicConfig(level=logging.WARN)
logging.getLogger('nshm_toshi_client.toshi_client_base').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)
logging.getLogger('botocore').setLevel(logging.INFO)
logging.getLogger('gql.transport.requests').setLevel(logging.WARN)

formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
screen_handler = logging.StreamHandler(stream=sys.stdout)
screen_handler.setFormatter(formatter)
file_handler = logging.FileHandler('slt.log')
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(formatter)
log.addHandler(screen_handler)
log.addHandler(file_handler)


@click.group()
def slt():
    """Inspect and produce model source_logic_trees."""
    pass


@slt.command(name='ls')
@click.option('-l', '--long', is_flag=True, help="also print the model title.")
def cli_ls(long):
    """List the available model versions."""

    for version in all_model_versions():
        if not long:
            click.echo(version)
            continue
        model = nzshm_model.get_model_version(version)
        click.echo(f"{model.version} `{model.title}`")


@slt.command(name='to_json')
@click.argument('model_id')
def cli_model_as_json(model_id: str):
    """Get the model in json form."""
    model = get_model_version(model_id)
    j = json.dumps(dataclasses.asdict(model.source_logic_tree), indent=4)
    click.echo(j)


@slt.command(name='hash_sources')
@click.argument('model_id')
@click.option('-o', '--outfile', type=click.File('w'))
def cli_model_source_hashes(model_id: str, outfile: io.FileIO):
    """Dump the sources with hashes form the given MODEL."""
    model = get_model_version(model_id)
    registry = branch_registry.BranchRegistry()
    for branch_set in model.source_logic_tree.branch_sets:
        for branch in branch_set.branches:
            entry = branch_registry.BranchRegistryEntry(identity=branch.registry_identity, extra=str(branch.tag))
            registry.add(entry)
            click.echo(entry)
    if outfile:
        registry.save(outfile)


@slt.command(name='hash_gmms')
@click.argument('model_id')
@click.option('-o', '--outfile', type=click.File('w'))
def cli_model_gmm_hashes(model_id: str, outfile: io.FileIO):
    """Dump the gmm branches with hashes."""
    model = get_model_version(model_id)
    registry = branch_registry.BranchRegistry()
    for branch_set in model.gmm_logic_tree.branch_sets:
        # print(dir(branch_set))
        for branch in branch_set.branches:
            entry = branch_registry.BranchRegistryEntry(branch.registry_identity)
            registry.add(entry)
            click.echo(entry)
    if outfile:
        registry.save(outfile)


# @slt.command(name='from_config')
# @click.argument('config_path')
# @click.argument('version')
# @click.argument('title')
# @click.option('-R', '--resolve_toshi_ids', is_flag=True)
# def cli_from_config(config_path, version, title, resolve_toshi_ids):
#     """Convert a python config file at CONFIG_PATH to an SLT model. Both VERSION and TITLE are required.

#     is this still required? Looks like it relates to old v1 SLTs
#     """

#     slt = from_config(config_path, version, title)

#     if resolve_toshi_ids:
#         slt = resolve_toshi_source_ids(slt)  # get new slt with toshi_ids
#     j = json.dumps(dataclasses.asdict(slt), indent=4)
#     click.echo(j)


# @slt.command(name='spec')
# @click.argument('model_id')
# @click.option('-B', '--build', is_flag=True)
# def cli_model_spec(model_id, build):
#     """Get a model specification by MODEL_ID."""
#     model = get_model_version(model_id)
#     if not model:
#         click.echo(f'Oops, the model "{model_id}" was not found', err=True)
#         return False
#     slt = model.source_logic_tree()
#     j = json.dumps(dataclasses.asdict(slt.derive_spec()), indent=4)
#     click.echo(j)


if __name__ == '__main__':
    slt()  # pragma: no cover



import pytest

from nzshm_model import get_model_version

CURRENT_VERSION = "NSHM_v1.0.4"


@pytest.fixture(scope='function')
def current_version():
    return CURRENT_VERSION


@pytest.fixture
def current_model(current_version):
    return get_model_version(current_version)



import json
from pathlib import Path

import pytest

from nzshm_model.logic_tree import InversionSource, SourceLogicTree


@pytest.fixture()
def slt_dict():
    slt_json_path = Path(__file__).parent / 'fixtures' / 'slt_config_sample.json'
    with slt_json_path.open() as slt_file:
        return json.load(slt_file)


def test_slt_from_config(slt_dict):
    slt = SourceLogicTree.from_dict(slt_dict)
    assert len(list(slt.composite_branches)) == 4
    assert len(list(slt._composite_branches())) == 4 * 2

    assert slt.branch_sets[2].branches[0].branch_id == ""
    assert slt.branch_sets[2].short_name == ""
    assert slt.branch_sets[0].branches[0].branch_id == "PUY1"
    assert slt.branch_sets[1].short_name == "HIK"
    assert slt.branch_sets[0].branches[1].weight == 0.8
    assert len(slt.branch_sets[0].branches[0].sources) == 4
    assert isinstance(slt.branch_sets[0].branches[0].sources[2], InversionSource)
    assert slt.branch_sets[1].branches[2].sources[0].nrml_id == "MNO"
    assert slt.correlations.correlation_groups[0].primary_branch == slt.branch_sets[1].branches[0]
    assert slt.correlations.correlation_groups[0].associated_branches == [slt.branch_sets[0].branches[0]]


def test_slt_from_config_errors0(slt_dict):
    # branch weights of branch set must sum to 1.0
    slt_dict["branch_sets"][0]["branches"][0]["weight"] = 0.3
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors1(slt_dict):
    # source type and member type should match
    # NB: the type member is not in the documentation, but a user could set it to the wrong value mistakenly
    slt_dict["branch_sets"][1]["branches"][2]["sources"][0]["type"] = "inversion"
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors2(slt_dict):
    # branches named in correlations must exist in the logic tree
    slt_dict["correlations"][0][0] = ["HIK:HIK10"]
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors3(slt_dict):
    # cannot repeat primary branch in correlations
    slt_dict["correlations"][1][0] = slt_dict["correlations"][0][0]
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors4(slt_dict):
    # every branch must have at least one source
    # slt_dict["branch_sets"][1][3]["sources"] = []
    del slt_dict["branch_sets"][1]["branches"][3]["sources"]
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors5(slt_dict):
    # don't accept duplicate branch_set.short_name:branch.branch_id
    slt_dict["branch_sets"][1]["short_name"] = "PUY"
    slt_dict["branch_sets"][1]["branches"][1]["branch_id"] = "PUY1"
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors6(slt_dict):
    # correlations must name branches in the LT
    slt_dict["correlations"][0][0] = "HIK:HIK10"
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors7(slt_dict):
    # correlations must name branches in the LT
    slt_dict["correlations"][0][0] = 1
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)


def test_slt_from_config_errors8(slt_dict):
    # correlations must name branches in the LT
    slt_dict["correlations"][0][0] = "HIKHIK1"
    with pytest.raises(ValueError) as value_error:
        SourceLogicTree.from_dict(slt_dict)
    print(value_error.value)



"""
NshmModel class describes a complete National Seismic Hazard Model.
"""

import importlib.resources as resources
import json
from pathlib import Path
from typing import Any, Dict, Generic, Iterator, List, Optional, Type, Union

from nzshm_model.logic_tree import GMCMLogicTree, SourceBranchSet, SourceLogicTree
from nzshm_model.logic_tree.source_logic_tree import SourceLogicTreeV1
from nzshm_model.model_versions import versions
from nzshm_model.psha_adapter import ModelPshaAdapterInterface
from nzshm_model.psha_adapter.hazard_config_factory import hazard_config_class_factory

from .psha_adapter.hazard_config import HazardConfig, HazardConfigType

RESOURCES_PATH = resources.files('nzshm_model.resources')

SLT_SOURCE_PATH = RESOURCES_PATH / "SRM_JSON"
GMM_JSON_SOURCE_PATH = RESOURCES_PATH / "GMM_JSON"
GMM_SOURCE_PATH = RESOURCES_PATH / "GMM_LTs"
HAZARD_CONFIG_PATH = RESOURCES_PATH / "HAZARD_CONFIG_JSON"


class NshmModel(Generic[HazardConfigType]):
    """
    An NshmModel instance represents a complete National Seismic Hazard Model version.
    """

    def __init__(
        self,
        version: str,
        title: str,
        source_logic_tree: SourceLogicTree,
        gmcm_logic_tree: GMCMLogicTree,
        hazard_config: HazardConfig,
    ):
        """
        Arguments:
            version: Describes version of the model being developed / used.
            tite: A title for the model.
            source_logic_tree: The seismicity rate model (SRM) logic tree.
            gmcm_logic_tree: The ground motion characterization model (GMCM) logic tree.
            hazard_config: The hazard engine (calculator) configuration.
        """
        self.version = version
        self.title = title
        self.hazard_config = hazard_config
        self.source_logic_tree = source_logic_tree
        self.gmm_logic_tree = gmcm_logic_tree

    @classmethod
    def from_files(
        cls,
        version: str,
        title: str,
        slt_json: Union[str, Path],
        gmm_json: Union[str, Path],
        hazard_config_json: Union[str, Path],
    ) -> 'NshmModel[HazardConfigType]':
        """
        Create a new NshmModel instance from files.

        NB library users will typically never use this, rather they will obtain a model instance
        using static method: `get_model_version`.
        """

        # backwards compatatilbity for v1 SourceLogicTree
        # v1 is not versioned
        data = NshmModel._slt_data_from_file(slt_json)
        if data.get("logic_tree_version") is None:
            source_logic_tree = NshmModel._source_logic_tree_from_v1_json(slt_json)

        source_logic_tree = SourceLogicTree.from_json(slt_json)
        gmcm_logic_tree = GMCMLogicTree.from_json(gmm_json)
        HazardConfigClass = hazard_config_class_factory.get_hazard_config_class_from_file(hazard_config_json)
        hazard_config = HazardConfigClass.from_json(hazard_config_json)
        return cls(version, title, source_logic_tree, gmcm_logic_tree, hazard_config)

    @staticmethod
    def _slt_data_from_file(filepath: Union[str, Path]) -> Dict[Any, Any]:
        with Path(filepath).open('r') as jsonfile:
            data = json.load(jsonfile)
        return data

    @staticmethod
    def _source_logic_tree_from_v1_json(filepath: Union[str, Path]) -> SourceLogicTree:
        """
        Create a SourceLogicTree from an old v1 json file; convert to new type.

        Arguments:
            filepath: the path to the json file specifying the source logic tree

        Returns:
            a source logic tree

        """
        return SourceLogicTree.from_source_logic_tree(SourceLogicTreeV1.from_json(filepath))

    @classmethod
    def get_model_version(cls, version: str) -> 'NshmModel':
        """
        Retrieve an existing model by its specific version

        Examples:
            >>> from nzshm_model import NshmModel
            >>> model = NshmModel.get_model_version("NSHM_v1.0.4")
            >>> print(model.title)
            >>>
            NSHM version 1.0.4, corrected fault geometry

        Parameters:
            version: The unique identifier for the model version.

        Raises:
            ValueError: when the version does not exist.

        Returns:
            the model instance.
        """

        model_args_factory = versions.get(version)
        if not model_args_factory:
            raise ValueError(f"{version} is not a valid model version.")

        model_args = model_args_factory()
        model_args['slt_json'] = SLT_SOURCE_PATH / model_args['slt_json']
        model_args['gmm_json'] = GMM_JSON_SOURCE_PATH / model_args['gmm_json']
        model_args['hazard_config_json'] = HAZARD_CONFIG_PATH / model_args['hazard_config_json']
        return cls.from_files(**model_args)

    def get_source_branch_sets(self, short_names: Union[List[str], str, None] = None) -> Iterator['SourceBranchSet']:
        """
        get an iterator for the SourceBranchSets matching the specified branch set(s)

        Examples:
            >>>  model = get_model_version("NSHM_v1.0.4")
            >>>  for branch_set in model.get_source_branch_sets(['CRU', 'PUY']):
                    print(branch_set.short_name, len(branch_set.branches))
            >>>
            CRU 36
            PUY 3

        Parameters:
            short_names: list of short_names for branch_set(s) (eg. 'HIK', 'CRU', 'PUY', 'SLAB')

        Raises:
            ValueError: when a branch short_name is not found.

        Yields:
            iterator of branch_set objects
        """
        if isinstance(short_names, str):
            list_short_names: List[str] = [short_names]
        else:
            list_short_names = short_names if short_names is not None else []

        if not list_short_names:  # User passes either an empty list or None
            for branch_set in self.source_logic_tree.branch_sets:
                yield branch_set
        else:
            # user has passes a list of short_names
            # check all the names are valid:
            for short_name in list_short_names:
                try:
                    for b in filter(lambda item: item.short_name == short_name, self.source_logic_tree.branch_sets):
                        yield (b)
                except StopIteration:
                    raise ValueError("The branch " + short_name + " was not found.")

    def psha_adapter(
        self, provider: Type[ModelPshaAdapterInterface], **kwargs: Optional[Dict]
    ) -> "ModelPshaAdapterInterface":
        """get a PSHA adapter for this instance.

        Arguments:
            provider: the adapter class
            **kwargs: additional arguments required by the provider class

        Returns:
            a PSHA Adapter instance
        """
        return provider(target=self)



#!python3 hazard_config.py
"""
Module supporting managed openquake configuration files

 - uses https://docs.python.org/3/library/configparser.html to do the heavy lifting
 - migrated from runzi /runzi/execute/openquake/util/oq_hazard_config.py

"""

import ast
import configparser
import copy
import json
import math
from itertools import chain
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, TextIO, Tuple, Type, Union, cast

from nzshm_common import CodedLocation

from nzshm_model.psha_adapter.hazard_config import HazardConfig, HazardConfigType

from .hazard_config_compat import check_invariants, compatible_hash_digest


# the z1pt0 and z2pt5 functions have been ported from openquake (openquake.hazardlib.site) to avoid incompatibility
# as the openquake API changes
def calculate_z1pt0(vs30: float) -> float:
    """
    Calculates z1.0 (depth to 1.0 km/s velocity horizon) in m. Assumes the California/global
    constants, not the Japan specific ones.

    Ref: Chiou, B. S.-J. and Youngs, R. R., 2014. 'Update of the Chiou and Youngs NGA model for the
    average horizontal component of peak ground motion and response spectra.' Earthquake Spectra,
    30(3), pp.1117–1153.

    Arguments:
        vs30: time averaged shear wave velocity from 0 to 30m depth (m/s)

    Returns:
        depth to 1.0 km/s velocity horizon in m
    """

    c1_glo = 571**4.0
    c2_glo = 1360.0**4.0
    return math.exp((-7.15 / 4.0) * math.log((vs30**4 + c1_glo) / (c2_glo + c1_glo)))


def calculate_z2pt5(vs30: float) -> float:
    """
    Calculates z2.5 (depth to 2.5 km/s velocity horizon) in km. Assumes the California constants,
    not the Japan specific ones.

    Ref: Campbell, K.W. & Bozorgnia, Y., 2014. 'NGA-West2 ground motion model for the average
    horizontal components of PGA, PGV, and 5pct damped linear acceleration response spectra.'
    Earthquake Spectra, 30(3), pp.1087–1114.

    Arguments:
        vs30: time averaged shear wave velocity from 0 to 30m depth (m/s)

    Returns:
        depth to 2.5 km/s velocity horizon in km
    """
    c1_glo = 7.089
    c2_glo = -1.144
    return math.exp(c1_glo + math.log(vs30) * c2_glo)


class OpenquakeConfig(HazardConfig):
    """Helper class to manage openquake configuration files.

    Examples:
        >>> from nzshm_model.psha_adapter.openquake import (
                OpenquakeConfig,\\
                DEFAULT_HAZARD_CONFIG\\
            )
        ...
        >>> oq_config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)\\
        >>>        .set_description("Hello openquake")\\
        >>>        .set_site_filepath("./sites.csv")\\
        >>>        .set_source_logic_tree_file(str(sources_filepath))\\
        >>>        .set_gsim_logic_tree_file("./gsim_model.xml")\\
        >>>        .set_vs30(750)
        ...
        >>> oq_config.config.get('general', 'description')
        ... Hello openquake
    """

    hazard_type = "openquake"

    def __init__(self, default_config: Union[configparser.ConfigParser, Dict, None] = None):

        self._site_parameters: Optional[Dict[str, Tuple]] = None
        self._locations: Optional[Tuple[CodedLocation]] = None

        if isinstance(default_config, configparser.ConfigParser):
            self.config = copy.deepcopy(default_config)
            return

        self.config = configparser.ConfigParser()
        if isinstance(default_config, dict):
            self.config.read_dict(default_config)
        return

    def __eq__(self, other) -> bool:
        if isinstance(other, OpenquakeConfig):
            return self.config == other.config
        return NotImplemented

    def is_complete(self) -> bool:
        return bool(self.get_iml() or self.get_iml_disagg())

    def _config_to_dict(self) -> Dict[str, Any]:
        data: Dict[str, Any] = dict()
        for section in self.config.sections():
            data[section] = dict()
            for k, v in self.config[section].items():
                data[section][k] = v
        return data

    def _locations_to_strs(self) -> List[str]:
        if not self.locations:
            return []
        return [loc.code for loc in self.locations]

    def to_dict(self) -> Dict[str, Any]:
        config_dict: Dict[str, Any] = dict(config=self._config_to_dict())
        config_dict['locations'] = self._locations_to_strs()
        config_dict['site_parameters'] = self._site_parameters
        config_dict['hazard_type'] = self.hazard_type
        return config_dict

    @classmethod
    def from_dict(cls: Type[HazardConfigType], data: Dict) -> 'OpenquakeConfig':
        site_parameters = data.pop('site_parameters')
        locations = data.pop('locations')
        hazard_config = cast('OpenquakeConfig', cls(data['config']))
        if site_parameters:
            hazard_config._site_parameters = hazard_config._deserialze_site_params(site_parameters)
        if locations:
            hazard_config._locations = hazard_config._deserialize_locations(locations)

        return hazard_config

    @classmethod
    def from_json(cls: Type[HazardConfigType], file_path: Union[Path, str]) -> 'OpenquakeConfig':
        with Path(file_path).open('r') as jsonfile:
            data = json.load(jsonfile)
        return cast('OpenquakeConfig', cls.from_dict(data))

    @staticmethod
    def _deserialze_site_params(site_parameters):
        data = dict()
        for k, v in site_parameters.items():
            data[k] = tuple(v)
        return data

    @staticmethod
    def _deserialize_locations(locations):
        # check that all coordinates have the same resolution
        def get_resolution(x):
            return 10 ** -(len(x) - x.find('.') - 1)

        all_coords = list(chain.from_iterable([loc.split('~') for loc in locations]))
        if len(set(map(get_resolution, all_coords))) != 1:
            raise Exception("not all coordinates have the same resolution")

        resolution = get_resolution(all_coords[0])
        ll_pairs = [[float(ll_str) for ll_str in loc.split('~')] for loc in locations]
        return tuple([CodedLocation.from_tuple(loc, resolution) for loc in ll_pairs])

    @staticmethod
    def read_file(config_file: TextIO) -> 'OpenquakeConfig':
        """produce a OpenquakeConfig from a file-like object

        Arguments:
            config_file: An file-like object with valid .ini contents.

        Returns:
            A new OpenquakeConfig instance.
        """
        config = configparser.ConfigParser()
        config.read_file(config_file)
        return OpenquakeConfig(config)

    def set_source_logic_tree_file(self, source_lt_filepath: Union[str, Path]) -> 'OpenquakeConfig':
        """setter for source_model file

        Arguments:
            source_lt_filepath: The path to the source model file.

        Returns:
            The OpenquakeConfig instance.
        """
        self.set_parameter("calculation", "source_model_logic_tree_file", str(source_lt_filepath))
        return self

    def set_parameter(self, section: str, key: str, value: Any) -> 'OpenquakeConfig':
        """a setter for arbitrary values

        Arguments:
            section: The config table name eg.[site_params].
            key: The key name.
            value: The value to set.

        Returns:
            The OpenquakeConfig instance.
        """
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, key, str(value))
        return self

    def get_parameter(self, section: str, key: str) -> Optional[str]:
        """A getter for arbitrary values.

        Arguments:
            section: The config table name eg.[site_params].
            key: The key name.

        Returns:
            The value for section and key or None if the entry does not exist.
        """
        if self.config.has_option(section, key):
            return self.config.get(section, key)
        return None

    def unset_parameter(self, section, key) -> 'OpenquakeConfig':
        """remove a table entry

        Arguments:
            section: The config table name eg.[site_params].
            key: The key name.

        Returns:
            The OpenquakeConfig instance.
        """
        if section in self.config:
            self.config.remove_option(section, key)
        return self

    def set_maximum_distance(self, value: dict[str, int]) -> 'OpenquakeConfig':
        """Set the maximum distance, which is a dictionary

        e.g. `{'Active Shallow Crust': 300.0, 'Volcanic': 300, 'Subduction Interface': 400, 'default': 400}`

        Arguments:
            value: Mapping of trt_names and distances.

        Returns:
            The OpenquakeConfig instance.
        """
        value_new = {}
        for trt, dist in value.items():
            if isinstance(dist, str):
                value_new[trt] = [tuple(dm) for dm in ast.literal_eval(dist)]
            else:
                value_new[trt] = dist
        self.config.set("calculation", "maximum_distance", str(value_new))
        return self

    def set_sites(self, locations: Iterable[CodedLocation], **site_parameters) -> 'OpenquakeConfig':
        """Setter for site_model file.

        If a vs30 values are specified, but a uniform vs30 has already been set a ValueError will be raised.

        Arguments:
            locations: The surface locations of the sites.
            kwargs: Additional site parameters to include in the OpenQuake site file.
                    Argument names will be used in the site file header. All entries
                    must be a sequence of the same length as locations.  See
                    https://docs.openquake.org/oq-engine/manual/latest/user-guide/inputs/site-model-inputs.html
                    for a list of valid site parameters.

        Returns:
            The OpenquakeConfig instance.
        """

        if 'vs30' in site_parameters and self.config.has_option('site_params', 'reference_vs30_value'):
            raise KeyError(
                "Cannot set site specific vs30, z1.0, or, z2.5: configuration specifies uniform site conditions."
            )
        if 'z1pt0' in site_parameters and self.config.has_option('site_params', 'reference_depth_to_1pt0km_per_sec'):
            raise KeyError(
                "Cannot set site specific vs30, z1.0, or, z2.5: configuration specifies uniform site conditions."
            )
        if 'z2pt5' in site_parameters and self.config.has_option('site_params', 'reference_depth_to_2pt5km_per_sec'):
            raise KeyError(
                "Cannot set site specific vs30, z1.0, or, z2.5: configuration specifies uniform site conditions."
            )

        self._site_parameters = {}
        locations = tuple(locations)
        for k, v in site_parameters.items():
            values = tuple(v)
            if not isinstance(v, Iterable):
                raise TypeError("all keyword arguments must be iterable type")
            if not len(values) == len(locations):
                raise ValueError("all keyword arguments must have the same number of elements as locations")
            self._site_parameters[k] = values

        self._locations = locations
        return self

    def set_site_filepath(self, site_file: Union[str, Path]) -> 'OpenquakeConfig':
        """
        Set the path to the site_model_file.
        """

        self.set_parameter('site_params', 'site_model_file', str(site_file))
        return self

    def get_site_filepath(self) -> Optional[Path]:
        value = self.get_parameter('site_params', 'site_model_file')
        return Path(value) if value else None

    @property
    def locations(self) -> Optional[Tuple[CodedLocation]]:
        return self._locations

    @property
    def site_parameters(self) -> Optional[Dict[str, tuple]]:
        return self._site_parameters

    def get_iml(self) -> Optional[Tuple[List[str], List[float]]]:
        """
        Get the intensity measure types and levels. Returns None if not set.

        Returns:
            a tuple of (IMTs, IMTLs) where IMTs is a list of intensity measure types and
            IMTLs is a list of intensity measure levels
        """

        value = self.get_parameter('calculation', 'intensity_measure_types_and_levels')
        if not value:
            return None

        imls = ast.literal_eval(value)
        imts = list(imls.keys())
        imtls = list([float(imtl) for imtl in next(iter(imls.values()))])

        return imts, imtls

    def get_iml_disagg(self) -> Optional[Tuple[str, float]]:
        """Get the intensity measure type and level for the disaggregation. Returns None if not set.

        Returns:
            a tuple of (IMT, and IMTL) where IMT is a intensity measure type and IMTL is an intensity measure level
        """
        value = self.get_parameter('disagg', 'iml_disagg')
        if not value:
            return None

        iml_imtl = ast.literal_eval(value)
        return list(iml_imtl.items())[0]

    # TODO disagg configs might warrant a separate class, and separate defaults ??
    def set_disagg_site_model(self) -> 'OpenquakeConfig':
        raise NotImplementedError()
        # self.set_parameter('site_params', 'site_model_file', 'site.csv')
        return self

    def set_disagg_site(self, lat, lon) -> 'OpenquakeConfig':
        raise NotImplementedError()
        # self.set_parameter('site_params', 'sites', f'{lon} {lat}')
        return self

    def set_iml_disagg(self, imt: str, level: float) -> 'OpenquakeConfig':
        self.set_parameter('disagg', 'iml_disagg', str({imt: level}))
        return self

    def clear_iml(self) -> 'OpenquakeConfig':
        """Remove intensity_measure_types_and_levels.

        Returns:
            The OpenquakeConfig instance.
        """
        self.unset_parameter('calculation', 'intensity_measure_types_and_levels')
        return self

    def set_iml(self, measures: List[str], levels: List[float]) -> 'OpenquakeConfig':
        """Setter for intensity_measure_types_and_levels

        Sets the same levels for all intensity measures.

        Arguments:
            measures: The IML types e.g `['PGA', 'SA(0.5)', ...].
            levels: The IML levels as floats  e.g. [0.01, 0.02, 0.04, ...].

        Returns:
            The OpenquakeConfig instance.
        """
        self.clear_iml()
        new_iml = '{'
        for m in measures:
            new_iml += f'"{m}": {str(levels)}, '
        new_iml += '}'

        if not self.config.has_section('calculation'):
            self.config.add_section('calculation')
        self.config['calculation']['intensity_measure_types_and_levels'] = new_iml
        return self

    def unset_uniform_site_params(self) -> 'OpenquakeConfig':
        """
        Remove the uniform site parameters from the configuration. This will unset the values for
        'reference_vs30_type', 'reference_vs30_value', 'reference_depth_to_1pt0km_per_sec', and
        'reference_depth_to_2pt5km_per_sec' in the 'site_params' section.
        """
        if not self.config.has_section('site_params'):
            return self

        sect = self.config['site_params']
        for setting in [
            'reference_vs30_type',
            'reference_vs30_value',
            'reference_depth_to_1pt0km_per_sec',
            'reference_depth_to_2pt5km_per_sec',
        ]:
            sect.pop(setting, None)

        return self

    def set_uniform_site_params(
        self, vs30: float, z1pt0: Optional[float] = None, z2pt5: Optional[float] = None
    ) -> 'OpenquakeConfig':
        """
        Setter for vs30, z1.0, and z2.5 site parameters.

        This will set the vs30, z1.0, and z2.5 site parameters for all sites. If z1pt0 and/or z2pt5
        are not specified they will be calculated from vs30. z1.0 is caculated using Chiou & Youngs
        (2014) California model and z2.5 is caclualted using Campbell & Bozorgnia 2014 NGA-West2 model.

        Arguments:
            vs30: the desired vs30
            z1pt0: the desired z1.0 depth in m
            z2pt5: the desired z2.5 depth in km

        Returns:
            The OpenquakeConfig instance.

        References:
            Campbell, K.W. & Bozorgnia, Y., 2014.
            'NGA-West2 ground motion model for the average horizontal components of
            PGA, PGV, and 5pct damped linear acceleration response spectra.' Earthquake Spectra,
            30(3), pp.1087–1114.

            Chiou, Brian & Youngs, Robert. (2014).
            'Update of the Chiou and Youngs NGA Model for the Average Horizontal Component of Peak
            Ground Motion and Response Spectra.' Earthquake Spectra. 30. 1117-1153.
        """

        if (site_parameters := (self.site_parameters)) and 'vs30' in site_parameters:
            raise KeyError("vs30 is already set as a site specific parameter")

        # clean up old settings
        self.unset_uniform_site_params()
        if not self.config.has_section('site_params'):
            self.config.add_section('site_params')
        sect = self.config['site_params']

        sect['reference_vs30_type'] = 'measured'
        sect['reference_vs30_value'] = str(vs30)

        if z1pt0:
            sect['reference_depth_to_1pt0km_per_sec'] = str(round(z1pt0, 0))
        elif 'calculate_z1pt0' in globals():
            sect['reference_depth_to_1pt0km_per_sec'] = str(round(calculate_z1pt0(vs30), 0))

        if z2pt5:
            sect['reference_depth_to_2pt5km_per_sec'] = str(round(z2pt5, 1))
        elif 'calculate_z2pt5' in globals():
            sect['reference_depth_to_2pt5km_per_sec'] = str(round(calculate_z2pt5(vs30), 1))

        return self

    def get_uniform_site_params(self) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        """
        The uniform site parameters of the model. Returns None if not set.

        Returns:
            (vs30, z1pt0, z2pt5) where vs30 is the vs30 applied to all sites, z1pt0 is the z1.0 reference
            depth in m, and z2pt5 is the z2.5 reference depth in km.
        """

        if not self.config.has_section('site_params'):
            return None, None, None

        value = self.get_parameter('site_params', 'reference_vs30_value')
        vs30 = float(value) if value else None

        value = self.config.get('site_params', 'reference_depth_to_1pt0km_per_sec', fallback=None)
        z1pt0 = float(value) if value else None

        value = self.config.get('site_params', 'reference_depth_to_2pt5km_per_sec', fallback=None)
        z2pt5 = float(value) if value else None

        return vs30, z1pt0, z2pt5

    def set_gsim_logic_tree_file(self, filepath: Union[str, Path]) -> 'OpenquakeConfig':
        """Setter for ground motion model file.

        Arguments:
            filepath: The path to the ground motion model file.

        Returns:
            The OpenquakeConfig instance.
        """
        self.set_parameter('calculation', 'gsim_logic_tree_file', str(filepath))
        return self

    def set_description(self, description) -> 'OpenquakeConfig':
        self.set_parameter('general', 'description', description)
        return self

    def write(self, tofile: TextIO) -> None:
        """Write the OpenquakeConfig to a file-like object.

        Arguments:
            tofile: A file-like object.
        """
        self.config.write(tofile)

    def compatible_hash_digest(self) -> str:
        """Get a shake_256 hash digest for the compatablity config.

        We want to ensure that, for this config:

         - all the invariant entries exist
         - entries that will not break calcluation compatibility are ignored

        Returns:
            The 12 character hash_digest.
        """
        check_invariants(self.config)
        return compatible_hash_digest(self.config)



from pathlib import Path

import pytest

from nzshm_model import get_model_version
from nzshm_model.logic_tree import GMCMLogicTree
from nzshm_model.psha_adapter.openquake import NrmlDocument
from nzshm_model.psha_adapter.openquake.simple_nrml import OpenquakeGMCMPshaAdapter

gmcm_nrml_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example.xml'
gmcm_json_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example.json'

adapter = get_model_version('NSHM_v1.0.4').gmm_logic_tree.psha_adapter(OpenquakeGMCMPshaAdapter)
gmcm_logic_tree_fromxml = adapter.gmcm_logic_tree_from_xml(gmcm_nrml_filepath)  # type: ignore
gmcm_logic_tree_fromjson = GMCMLogicTree.from_json(gmcm_json_filepath)


def test_gmcm_logic_tree_from_xml():
    doc = NrmlDocument.from_xml_file(gmcm_nrml_filepath)
    gmcm_logic_tree = gmcm_logic_tree_fromxml

    assert len(gmcm_logic_tree.branch_sets) == len(doc.logic_trees[0].branch_sets)
    # for i in range(len(gmcm_logic_tree.branch_sets)):
    #     assert (
    #         gmcm_logic_tree.branch_sets[i].tectonic_region_type
    #         == doc.logic_trees[0].branch_sets[i].applyToTectonicRegionType
    #     )

    assert len(gmcm_logic_tree.branch_sets[0].branches) == len(doc.logic_trees[0].branch_sets[0].branches)
    assert (
        gmcm_logic_tree.branch_sets[0].branches[0].weight
        == doc.logic_trees[0].branch_sets[0].branches[0].uncertainty_weight
    )
    assert gmcm_logic_tree.branch_sets[0].branches[0].gsim_name == 'Stafford2022'
    expected_args = dict(mu_branch='Upper')
    assert gmcm_logic_tree.branch_sets[0].branches[0].gsim_args == expected_args


def test_serialize_gmcm_logic_tree():
    lt_as_dict = gmcm_logic_tree_fromjson.to_dict()
    gmcm_logic_tree = GMCMLogicTree.from_dict(lt_as_dict)
    assert gmcm_logic_tree == gmcm_logic_tree_fromjson


@pytest.mark.parametrize("gmcm_logic_tree", [gmcm_logic_tree_fromxml, gmcm_logic_tree_fromjson])
def test_float_args(gmcm_logic_tree):
    def is_number(value):
        return value.replace("-", "").replace(".", "").replace("e", "").isnumeric()

    for branch_set in gmcm_logic_tree.branch_sets:
        for branch in branch_set.branches:
            for value in branch.gsim_args.values():
                if is_number(str(value)):
                    assert isinstance(value, float)



from collections import namedtuple
from unittest.mock import patch  # TODO: use pytest to patch?

import pytest

from nzshm_model.logic_tree.branch import Branch
from nzshm_model.logic_tree.correlation import Correlation, LogicTreeCorrelations
from nzshm_model.logic_tree.logic_tree_base import BranchSet, LogicTree

Fixtures = namedtuple(
    "Fixtures",
    """correlation1 correlation2 branchA1 branchA2 branchB1 branchB2
    branchsetA branchsetB logic_tree logic_tree_nocor logic_tree2""",
)


@pytest.fixture(scope='module')
@patch.multiple(LogicTree, __abstractmethods__=set())
@patch.multiple(LogicTreeCorrelations, __abstractmethods__=set())
@patch.multiple(Branch, __abstractmethods__=set())
def fixtures():
    branchA1 = Branch(branch_id="branchA1", weight=0.2)
    branchA2 = Branch(branch_id="branchA2", weight=0.2)
    branchA3 = Branch(branch_id="branchA3", weight=0.3)
    branchA4 = Branch(branch_id="branchA4", weight=0.3)
    branchB1 = Branch(branch_id="branchB1", weight=0.25)
    branchB2 = Branch(branch_id="branchB2", weight=0.75)
    branchC1 = Branch(branch_id="branchC1", weight=0.2)
    branchC2 = Branch(branch_id="branchC2", weight=0.8)

    branchsetA = BranchSet(short_name="A", long_name="branchsetA", branches=[branchA1, branchA2, branchA3, branchA4])
    branchsetB = BranchSet(short_name="B", long_name="branchsetB", branches=[branchB1, branchB2])
    branchsetC = BranchSet(short_name="C", long_name="branchsetC", branches=[branchC1, branchC2])

    correlation1 = Correlation(primary_branch=branchA1, associated_branches=[branchB1])
    correlation2 = Correlation(primary_branch=branchA2, associated_branches=[branchB2])
    correlation3 = Correlation(primary_branch=branchA3, associated_branches=[branchB1])
    correlation4 = Correlation(primary_branch=branchA4, associated_branches=[branchB2])
    correlations = LogicTreeCorrelations(correlation_groups=[correlation1, correlation2, correlation3, correlation4])

    logic_tree_nocor = LogicTree(title='logic_tree', branch_sets=[branchsetA, branchsetB, branchsetC])
    logic_tree = LogicTree(title='logic_tree', branch_sets=[branchsetA, branchsetB], correlations=correlations)
    logic_tree2 = LogicTree(
        title='logic_tree', branch_sets=[branchsetA, branchsetB, branchsetC], correlations=correlations
    )

    return Fixtures(
        correlation1=correlation1,
        correlation2=correlation2,
        branchA1=branchA1,
        branchA2=branchA2,
        branchB1=branchB1,
        branchB2=branchB2,
        branchsetA=branchsetA,
        branchsetB=branchsetB,
        logic_tree=logic_tree,
        logic_tree_nocor=logic_tree_nocor,
        logic_tree2=logic_tree2,
    )


def test_branchset(fixtures: Fixtures):
    # cannot create BranchSet if weights do not sum to 1.0
    with pytest.raises(ValueError):
        BranchSet(branches=[fixtures.branchA1, fixtures.branchB2])


def test_iterate_branchset(fixtures: Fixtures):
    for branch1, branch2 in zip(fixtures.branchsetA.branches, fixtures.branchsetA):
        assert branch1 == branch2


def test_correlation(fixtures: Fixtures):
    # we get the correct nubmer of total branches in a correlation
    assert len(fixtures.correlation1.all_branches) == 2


def test_check_correlations_validation(fixtures: Fixtures):
    # should not raise exeption
    LogicTreeCorrelations(correlation_groups=[fixtures.correlation1, fixtures.correlation2])

    # should raise exception same entry used twice
    with pytest.raises(ValueError):
        correlation2x = Correlation(primary_branch=fixtures.branchA1, associated_branches=[fixtures.branchB2])
        LogicTreeCorrelations(correlation_groups=[fixtures.correlation1, correlation2x])


def test__composite_branches(fixtures: Fixtures):

    assert len(list(fixtures.logic_tree_nocor.composite_branches)) == 4 * 2 * 2

    # branches not filtered by correlation
    assert len(list(fixtures.logic_tree._composite_branches())) == 4 * 2

    # branches filtered by correlation
    assert len(list(fixtures.logic_tree.composite_branches)) == 4

    # can correlate a subset of BranchSets
    assert len(list(fixtures.logic_tree2._composite_branches())) == 4 * 2 * 2

    assert len(list(fixtures.logic_tree2.composite_branches)) == 4 * 2


def test_composite_weights(fixtures: Fixtures):
    # weights sum to 1.0
    assert sum([branch.weight for branch in fixtures.logic_tree.composite_branches]) == pytest.approx(1.0)
    assert sum([branch.weight for branch in fixtures.logic_tree_nocor.composite_branches]) == pytest.approx(1.0)

    # weights are default from the primary branch
    for cg in fixtures.logic_tree.correlations.correlation_groups:
        assert cg.weight == cg.primary_branch.weight


def test_logic_tree_validation_of_corr(fixtures: Fixtures):
    # should not raise an execption
    correlation1 = Correlation(primary_branch=fixtures.branchA1, associated_branches=[fixtures.branchB1], weight=0.2)
    correlation2 = Correlation(primary_branch=fixtures.branchA2, associated_branches=[fixtures.branchB2], weight=0.2)
    correlations = LogicTreeCorrelations(correlation_groups=[correlation1, correlation2])
    fixtures.logic_tree.correlations = correlations

    # should not be able to set correlations with incorrect weights
    correlation1.weight = 1.0
    correlation2.weight = 1.0
    correlations = LogicTreeCorrelations(correlation_groups=[correlation1, correlation2])
    with pytest.raises(ValueError):
        fixtures.logic_tree.correlations = correlations



#! python test_logic_tree.py

# import dataclasses
# import itertools
from pathlib import Path

from nzshm_model.logic_tree import Correlation, LogicTreeCorrelations, SourceLogicTree

# import pytest


def test_v2_source_tree_from_json_no_correlations():
    config = Path(__file__).parent / 'fixtures' / 'source_logic_tree_sample_2.json'
    slt = SourceLogicTree.from_json(config)

    # without correlations there are 4 composite branches
    assert len(list(slt.composite_branches)) == 4

    cor1 = Correlation(
        primary_branch=slt.branch_sets[0].branches[0],
        associated_branches=[slt.branch_sets[1].branches[0]],
    )
    cor2 = Correlation(
        primary_branch=slt.branch_sets[0].branches[1],
        associated_branches=[slt.branch_sets[1].branches[1]],
    )
    slt.correlations = LogicTreeCorrelations([cor1, cor2])

    # with correlations there are 2 composite branches
    for branch in slt.composite_branches:
        print(branch)
        print('')
    assert len(list(slt.composite_branches)) == 2



#! nshm_v1_0_4.py
"""
This module defines a factory function for generating an NshmModel object for the model version NSHM_v1.0.4.
The dict returned by the function is passed as keyword arguments to the NshmModel constructor:

    >>> from nzshm_model import NshmModel
    >>> from nzshm_model.model_versions.nshm_v1_0_4 import model_args_factory
    >>> model = NshmModel(**model_args_factory())

NB library users will typically never use this, rather they will obtain a model instance
using function: `nzshm_model.get_model_version`.
"""

from typing import Any, Dict


def model_args_factory() -> Dict[str, Any]:
    return dict(
        version='NSHM_v1.0.4',
        title="NSHM version 1.0.4, corrected fault geometry",
        slt_json="nshm_v1.0.4_v2.json",
        gmm_json="gmcm_nshm_v1.0.4.json",
        hazard_config_json="oq_config_nshm_v1.0.4.json",
    )



from pathlib import Path

import pytest

from nzshm_model import all_model_versions, get_model_version
from nzshm_model.logic_tree import GMCMLogicTree, SourceLogicTree


@pytest.fixture(scope='module')
def branchset_names():
    return {
        'Active Shallow Crust': {'short_name': 'CRU', 'long_name': 'Crustal'},
        'Subduction Interface': {'short_name': 'INTER', 'long_name': 'Subduction Interface'},
        'Subduction Intraslab': {'short_name': 'SLAB', 'long_name': 'Subduction Intraslab'},
    }


source_json_files = ["nshm_v1.0.0_v2.json", "nshm_v1.0.4_v2.json"]


@pytest.mark.parametrize("model_version, json_filename", zip(all_model_versions(), source_json_files))
def test_json_source(model_version, json_filename):
    model = get_model_version(model_version)
    json_filepath = Path(__file__).parent / 'fixtures' / 'model_versions' / json_filename
    slt_expected = SourceLogicTree.from_json(json_filepath)
    slt_from_model = model.source_logic_tree
    assert slt_from_model == slt_expected


gmcm_json_files = ["gmcm_nshm_v1.0.0.json", "gmcm_nshm_v1.0.4.json"]


@pytest.mark.parametrize("model_version, json_filename", zip(all_model_versions(), gmcm_json_files))
def test_json_gmcm(model_version, json_filename):
    model = get_model_version(model_version)
    json_filepath = Path(__file__).parent / 'fixtures' / 'model_versions' / json_filename
    gmcm_expected = GMCMLogicTree.from_json(json_filepath)
    gmcm_from_model = model.gmm_logic_tree
    assert gmcm_from_model == gmcm_expected


@pytest.mark.parametrize("model_version", all_model_versions())
def test_brachset_names(model_version, branchset_names):
    model = get_model_version(model_version)
    gmcm = model.gmm_logic_tree
    for branch_set in gmcm.branch_sets:
        assert branch_set.short_name == branchset_names[branch_set.tectonic_region_type]['short_name']
        assert branch_set.long_name == branchset_names[branch_set.tectonic_region_type]['long_name']



from pathlib import Path

from nzshm_model.logic_tree import GMCMBranch, GMCMLogicTree

gmcm_json_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example.json'

gmcm_logic_tree = GMCMLogicTree.from_json(gmcm_json_filepath)


def test_filter_trt():
    gsim_args = {'epistemic': 'Central', 'modified_sigma': 'true'}
    # assert (
    #     len(
    #         [
    #             filt_branch
    #             for filt_branch in filter(
    #                 lambda obj: obj.branch_set.tectonic_region_type == 'Active Shallow Crust', gmcm_logic_tree
    #             )
    #         ]
    #     )
    #     == 21
    # )
    assert (
        len([filt_branch for filt_branch in filter(lambda obj: obj.gsim_name == 'Atkinson2022SInter', gmcm_logic_tree)])
        == 3
    )
    assert len([filt_branch for filt_branch in filter(lambda obj: obj.gsim_args == gsim_args, gmcm_logic_tree)]) == 3


def test_build_gmcmlt_from_filtered_gmcmlt():
    flt = (filt_branch for filt_branch in filter(lambda obj: obj.gsim_name == 'Atkinson2022SInter', gmcm_logic_tree))
    glt = GMCMLogicTree.from_branches(flt)
    print(glt)
    assert len(glt.branch_sets) == 1
    assert len(glt.branch_sets[0].branches) == 3
    assert (
        type(glt.branch_sets[0].branches[0]) is GMCMBranch
    )  # isinstance() not used to avoid true for inherited classes



import io

import pytest

from nzshm_model import branch_registry


@pytest.fixture(scope='module')
def gmm_csv_fixture():
    csv = """
hash_digest,identity,extra
380a95154af2,"Atkinson2022SInter(epistemic=Central, modified_sigma=true)",
957eb00fd580,"Atkinson2022SInter(epistemic=Lower, modified_sigma=true)",
772d4ab2272f,"NZNSHM2022_AbrahamsonGulerce2020SInter(region=GLO, sigma_mu_epsilon=1.28155)",
"""
    yield io.StringIO(csv)


@pytest.fixture(scope='module')
def sources_csv_fixture():
    csv = """
hash_digest,identity,extra
ef55f8757069,RmlsZToxMzA3MDc=|SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE1MDE=,"[dmgeodetic, tdFalse, bN[0.823, 2.7], C4.2, s0.66]"
"""  # noqa
    yield io.StringIO(csv)


class TestBranchRegistry:
    def test_init_sources_registry(self, sources_csv_fixture):
        # registry_file = open(sources_csv_fixture, 'r')
        sources_registry = branch_registry.BranchRegistry().load(sources_csv_fixture)
        assert len(sources_registry) == 1

    def test_init_gmm_registry(self, gmm_csv_fixture):
        gmm_registry = branch_registry.BranchRegistry().load(gmm_csv_fixture)
        assert len(gmm_registry) == 3

    def test_add_to_registry(self, gmm_csv_fixture):
        gmm_registry = branch_registry.BranchRegistry().load(gmm_csv_fixture)
        new_entry = branch_registry.BranchRegistryEntry(identity="SomeGMM")
        gmm_registry.add(new_entry)

        assert len(gmm_registry) == 4
        assert gmm_registry.get_by_hash(new_entry.hash_digest) == new_entry
        assert gmm_registry.get_by_identity(new_entry.identity) == new_entry

    def test_save_registry(self, gmm_csv_fixture):
        registry = branch_registry.BranchRegistry().load(gmm_csv_fixture)
        output_file = io.StringIO()
        registry.save(output_file)
        assert output_file.read() == gmm_csv_fixture.read()


class TestBranchRegistryEntry:
    def test_auto_digest(self):
        new_entry = branch_registry.BranchRegistryEntry("SomeGMM")
        assert new_entry.hash_digest == branch_registry.identity_digest("SomeGMM")
        assert new_entry.identity == "SomeGMM"

    def test_validate_user_digest_bad(self):
        with pytest.raises(ValueError):
            new_entry = branch_registry.BranchRegistryEntry("SomeGMM", hash_digest="ABC")
            assert new_entry

    def test_validate_user_digest_good(self):
        new_entry = branch_registry.BranchRegistryEntry(
            "SomeGMM", hash_digest=branch_registry.identity_digest("SomeGMM")
        )
        assert new_entry.hash_digest == branch_registry.identity_digest("SomeGMM")
        assert new_entry.identity == "SomeGMM"


class TestRegistryClass:
    def test_source_registry(self):
        registry = branch_registry.Registry()
        assert len(registry.source_registry) == 49
        entry = registry.source_registry.get_by_identity("RmlsZToxMzA3NTM=|SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE2NTc=")
        assert entry.hash_digest == "af9ec2b004d7"

    def test_source_registry_by_extra(self):
        """
        ef55f8757069,RmlsZToxMzA3MDc=|SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE1MDE=,
        "[dmgeodetic, tdFalse, bN[0.823, 2.7], C4.2, s0.66]"
        """
        registry = branch_registry.Registry()
        assert len(registry.source_registry) == 49
        entry = registry.source_registry.get_by_extra("[dmgeodetic, tdFalse, bN[0.823, 2.7], C4.2, s0.66]")
        assert entry.hash_digest == "ef55f8757069"
        assert entry.identity == "RmlsZToxMzA3MDc=|SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE1MDE="



"""
support for managing NSHM compatability in openquake caclulations
"""

import configparser
import copy
import hashlib
import io

DEFAULT_HAZARD_CONFIG: dict = dict(
    general=dict(
        random_seed=25,
        calculation_mode='classical',
        ps_grid_spacing=30,
    ),
    logic_tree=dict(
        number_of_logic_tree_samples=0,
    ),
    erf=dict(
        rupture_mesh_spacing=4,
        width_of_mfd_bin=0.1,
        complex_fault_mesh_spacing=10.0,
        area_source_discretization=10.0,
    ),
    site_params=dict(
        reference_vs30_type='measured',
    ),
    calculation=dict(
        investigation_time=1.0,
        truncation_level=4,
        maximum_distance={
            'Active Shallow Crust': [(4.0, 0), (5.0, 100.0), (6.0, 200.0), (9.5, 300.0)],
            'Subduction Interface': [(5.0, 0), (6.0, 200.0), (10, 500.0)],
            'Subduction Intraslab': [(5.0, 0), (6.0, 200.0), (10, 500.0)],
        },
    ),
    output=dict(
        individual_curves='true',
    ),
)

ENTRIES_IGNORED = [
    ('general', 'description'),
    ('general', 'random_seed'),
    ('site_params', 'sites_csv'),
    ('site_params', 'sites'),
    ('site_params', 'site_model_file'),
    ('site_params', 'reference_vs30_value'),
    ('site_params', 'reference_depth_to_1pt0km_per_sec'),  # aka z1
    ('site_params', 'reference_depth_to_2pt5km_per_sec'),  # aka z2.5
    ('erf', 'concurrent_tasks'),
    ('calculation', 'source_model_logic_tree_file'),
    ('calculation', 'gsim_logic_tree_file'),
    ('calculation', 'intensity_measure_types_and_levels'),
]

ENTRIES_INVARIANT = [
    ('general', 'calculation_mode', 'classical'),
    ('logic_tree', 'number_of_logic_tree_samples', '0'),
    ('site_params', 'reference_vs30_type', 'measured'),
    (
        'calculation',
        'investigation_time',
        '1.0',
    ),
    ('output', 'individual_curves', 'true'),
]


def check_invariants(config: configparser.ConfigParser) -> bool:
    """check a configuration has all the expected invariant entries"""
    for table, key, value in ENTRIES_INVARIANT:
        try:
            invariant = config.get(table, key)
        except configparser.NoOptionError:
            raise ValueError(f'Expected entry "[{table}]" "{key}"" with value {value} was not found.')
        if invariant == value:
            continue
        raise ValueError(
            f'Expected entry "[{table}]" "{key}"" with value {value} was not found. Got {invariant} instead.'
        )
    return True


def compatible_config(config: configparser.ConfigParser) -> configparser.ConfigParser:
    clean_config = copy.deepcopy(config)
    for table, key in ENTRIES_IGNORED:
        clean_config.remove_option(table, key)
    return clean_config


def compatible_hash_digest(config: configparser.ConfigParser, digest_len: int = 12) -> str:
    """return a 12 character hexdigest for the config"""

    if digest_len % 2 > 0:
        raise ValueError('len: {digest_len} is not divisible by 2')

    check_invariants(config)  # raises an error if not sucessful
    compat = compatible_config(config)
    value = io.StringIO()
    compat.write(value)
    return hashlib.shake_256(value.getvalue().encode()).hexdigest(int(digest_len / 2))



#! python test_logic_tree.py
import json
from pathlib import Path

import pytest

import nzshm_model.psha_adapter.openquake.simple_nrml
from nzshm_model.logic_tree import GMCMLogicTree, SourceLogicTree
from nzshm_model.psha_adapter.openquake import OpenquakeGMCMPshaAdapter, OpenquakeModelPshaAdapter

FIXTURE_PATH = Path(__file__).parent.parent.parent / "fixtures"


@pytest.mark.skip("WIP")
def test_source_logic_tree_to_source_xml_basic():

    config = FIXTURE_PATH / 'source_logic_tree_sample_0.json'
    slt = SourceLogicTree.from_json(config)

    print(slt)
    target = FIXTURE_PATH  # noqa
    src_xml = slt.psha_adapter(provider=OpenquakeModelPshaAdapter).write_config(Path('/tmp/DEMO'))  # noqa
    # assert 0


@pytest.mark.skip("WIP")
def test_fetch_resources(monkeypatch, tmp_path):
    def mockreturn(file_id, destination):
        return file_id

    config = FIXTURE_PATH / 'source_logic_tree_sample_2.json'
    slt = SourceLogicTree.from_json(config)

    monkeypatch.setattr(nzshm_model.psha_adapter.openquake.simple_nrml, "fetch_toshi_source", mockreturn)
    result = [
        {'id': _id, 'path': filepath, 'um': um}
        for _id, filepath, um in slt.psha_adapter(provider=OpenquakeModelPshaAdapter).fetch_resources(tmp_path)
    ]

    with Path('test.json').open('w') as jsonfile:
        json.dump(result, jsonfile)


def test_gmcm_logic_tree_to_xml(tmp_path):
    gmcm_json_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example.json'

    gmcm_logic_tree = GMCMLogicTree.from_json(gmcm_json_filepath)
    adapter = gmcm_logic_tree.psha_adapter(OpenquakeGMCMPshaAdapter)
    xml_str = adapter.build_gmcm_xml()
    with (tmp_path / 'gmcm_lt.xml').open('w') as xmlfile:
        xmlfile.write(xml_str)
    gmcm_logic_tree_deserialized = adapter.gmcm_logic_tree_from_xml(tmp_path / 'gmcm_lt.xml')

    assert gmcm_logic_tree_deserialized == gmcm_logic_tree



import pytest

from nzshm_model import get_model_version
from nzshm_model.logic_tree.gmcm_logic_tree import GMCMBranch, GMCMBranchSet
from nzshm_model.logic_tree.source_logic_tree import SourceBranch, SourceBranchSet
from nzshm_model.logic_tree.source_logic_tree.logic_tree import DistributedSource


def test_source_trts():
    b1 = SourceBranch(
        tectonic_region_types=("a", "b"), branch_id="1", weight=0.5, sources=[DistributedSource(nrml_id="ABC")]
    )
    b2 = SourceBranch(
        tectonic_region_types=("a", "b"), branch_id="2", weight=0.5, sources=[DistributedSource(nrml_id="DEF")]
    )
    b3 = SourceBranch(
        tectonic_region_types=("c", "b"), branch_id="2", weight=0.5, sources=[DistributedSource(nrml_id="DEF")]
    )

    with pytest.raises(ValueError):
        sbs = SourceBranchSet(branches=(b1, b2, b3))

    sbs = SourceBranchSet(branches=(b1, b2))
    assert sbs.tectonic_region_types == ("a", "b")

    with pytest.raises(AttributeError):
        sbs.tectonic_region_types = ("c", "d")


def test_gmcm_trts():
    b1 = GMCMBranch(tectonic_region_type="a", branch_id="1", weight=0.5)
    b2 = GMCMBranch(tectonic_region_type="a", branch_id="2", weight=0.5)
    b3 = GMCMBranch(tectonic_region_type="b", branch_id="2", weight=0.5)

    with pytest.raises(ValueError):
        gbs = GMCMBranchSet(branches=(b1, b2, b3))

    gbs = GMCMBranchSet(branches=(b1, b2))
    assert gbs.tectonic_region_type == "a"

    with pytest.raises(AttributeError):
        gbs.tectonic_region_type = "c"


def test_model_trts():
    model = get_model_version('NSHM_v1.0.4')

    glt = model.gmm_logic_tree
    trts = ["Active Shallow Crust", "Subduction Interface", "Subduction Intraslab"]
    for branch_set, trt in zip(glt.branch_sets, trts):
        assert branch_set.tectonic_region_type == trt
        assert branch_set.branches[0].tectonic_region_type == trt

    slt = model.source_logic_tree
    trts = ["Subduction Interface", "Subduction Interface", "Active Shallow Crust", "Subduction Intraslab"]
    for branch_set, trt in zip(slt.branch_sets, trts):
        assert branch_set.tectonic_region_types == (trt,)
        assert branch_set.branches[0].tectonic_region_types == (trt,)



"""
This module defines the interface to be provided by a PshaAdapter implementation.
"""

import pathlib
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union


class PshaAdapterInterface(ABC):
    """
    Defines methods to be provided by a PSHA adapter class implementation.
    """

    @abstractmethod
    def __init__(self, target: Any):
        """
        Create a new PshaAdapterInterface object.

        Parameters:
            target: the object (e.g. SourceLogicTree) to be adapted to the specific hazard engine
        """
        pass

    @abstractmethod
    def write_config(self, *args, **kwargs) -> pathlib.Path:
        pass


class SourcePshaAdapterInterface(PshaAdapterInterface):
    """
    Defines methods to be provided by a PSHA source model adapter class implementation.
    """

    @abstractmethod
    def fetch_resources(self, cache_folder):
        """pull required data from api and store in target_folder"""
        pass

    @abstractmethod
    def unpack_resources(self, cache_folder: Union[pathlib.Path, str], target_folder: Union[pathlib.Path, str]):
        """unpack resources from cache_folder to target folder returning mapping of unpacked filepaths"""
        pass

    @abstractmethod
    def write_config(
        self,
        cache_folder: Union[pathlib.Path, str],
        target_folder: Union[pathlib.Path, str],
        resource_map: Optional[Dict[str, list[pathlib.Path]]] = None,
    ) -> pathlib.Path:
        """Build PSHA source input files"""
        pass


class GMCMPshaAdapterInterface(PshaAdapterInterface):
    """
    Defines methods to be provided by a PSHA ground motion model adapter class implementation.
    """

    @abstractmethod
    def write_config(self, target_folder: Union[pathlib.Path, str]) -> pathlib.Path:
        """Build PSHA GMCM input files"""
        pass


class ConfigPshaAdapterInterface(PshaAdapterInterface):
    """
    Defines methods to be provided by a PSHA ground motion model adapter class implementation.
    """

    @abstractmethod
    def write_config(self, target_folder: Union[pathlib.Path, str]) -> pathlib.Path:
        """Build PSHA calculation config input files"""
        pass


class ModelPshaAdapterInterface(PshaAdapterInterface):
    """
    Defines methods to be provided by a PSHA model adapter class implementation.
    """

    @abstractmethod
    def write_config(
        self,
        cache_folder: Union[pathlib.Path, str],
        target_folder: Union[pathlib.Path, str],
        resource_map: Optional[Dict[str, list[pathlib.Path]]] = None,
    ) -> pathlib.Path:
        """Build all PSHA input files"""
        pass



import configparser

import pytest

import nzshm_model as nm
from nzshm_model.psha_adapter.openquake.hazard_config_compat import DEFAULT_HAZARD_CONFIG


def test_list_all_models(current_version):
    assert current_version in nm.all_model_versions()


def test_mutable_model(current_version):
    model = nm.get_model_version(current_version)
    slt = model.source_logic_tree
    slt.branch_sets = [bs for bs in slt.branch_sets if bs.short_name == "CRU"]
    model.source_logic_tree.branch_sets = [bs for bs in model.source_logic_tree.branch_sets if bs.short_name == "CRU"]
    assert len(model.source_logic_tree.branch_sets) == len(slt.branch_sets)


class TestLoadModel:
    def test_load_model_104(self, current_version):
        model = nm.model.NshmModel.get_model_version(current_version)
        assert model

    def test_model_104_title(self, current_model):
        assert current_model.title == "NSHM version 1.0.4, corrected fault geometry"


class TestGetSourceBranchSets:
    def test_with_list(self, current_model):
        branch_sets = list(current_model.get_source_branch_sets(['CRU', 'PUY']))
        assert len(branch_sets) == 2
        assert branch_sets[0].short_name == 'CRU'
        assert branch_sets[1].short_name == 'PUY'

    def test_with_single(self, current_model):
        branch_sets = list(current_model.get_source_branch_sets('CRU'))
        assert len(branch_sets) == 1
        assert branch_sets[0].short_name == 'CRU'

    def test_with_null(self, current_model):
        assert len(list(current_model.get_source_branch_sets([]))) == 4
        assert len(list(current_model.get_source_branch_sets())) == 4

    def test_with_invalid_branch(self, current_model):
        with pytest.raises(StopIteration):
            next(current_model.get_source_branch_sets(['XXX']))


class TestConfig:
    def test_openquake_config(self, current_model):
        expected = configparser.ConfigParser()
        expected.read_dict(DEFAULT_HAZARD_CONFIG)
        assert current_model.hazard_config.config == expected


class TestGetNewModel:
    def test_get_new_model(self, current_version):
        new_model_title = "NEW MODEL TITLE"
        model1 = nm.get_model_version(current_version)
        model1.title = new_model_title

        model2 = nm.get_model_version(current_version)
        assert model1.title == new_model_title
        assert model2.title != new_model_title


# class TestGetSourceBranches:
#     def test_get_first_crustal_branch(self, model_104):
#         cru_branches = next(model_104.get_source_branch_sets('CRU')).branches
#         assert next(model_104.get_source_branches('CRU')) == cru_branches[0]

#     def test_invalid_branch_set(self, model_104):
#         with pytest.raises(StopIteration):
#             next(model_104.get_source_branches(['XXX']))



"""
tests for the OpenquakeConfiguration class
"""

import configparser
import io

# import tomli
import pytest
from nzshm_common.location import CodedLocation

import nzshm_model.psha_adapter.openquake.hazard_config as oqhc
from nzshm_model.psha_adapter.openquake.hazard_config import OpenquakeConfig
from nzshm_model.psha_adapter.openquake.hazard_config_compat import (
    DEFAULT_HAZARD_CONFIG,
    check_invariants,
    compatible_config,
    compatible_hash_digest,
)

measures = ['PGA', 'SA(0.5)']

min_expected = """
[general]
ps_grid_spacing = 30
calculation_mode = classical

[logic_tree]
number_of_logic_tree_samples = 0

[erf]
rupture_mesh_spacing = 5
width_of_mfd_bin = 0.1
complex_fault_mesh_spacing = 10.0
area_source_discretization = 10.0

[site_params]
sites_csv = grid-NZ-0.2-0.0003.nb1.nz34.csv
reference_vs30_type = measured
reference_vs30_value = 750
reference_depth_to_1pt0km_per_sec = 44.0 // z1_0
reference_depth_to_2pt5km_per_sec = 0.6  // Z2_5

[calculation]
source_model_logic_tree_file = ./sources/source_model.xml
gsim_logic_tree_file = ./NZ_NSHM_GMM_LT_final_EE.xml
investigation_time = 1.0

truncation_level = 4
maximum_distance = {'Active Shallow Crust': [(4.0, 0), (5.0, 100.0), (6.0, 200.0), (9.5, 300.0)],
                        'Subduction Interface': [(5.0, 0), (6.0, 200.0), (10, 500.0)],
                        'Subduction Intraslab': [(5.0, 0), (6.0, 200.0), (10, 500.0)]}

intensity_measure_types_and_levels  = {"PGA": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.1)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.2)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.3)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.4)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.5)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(0.7)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(1.0)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(1.5)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(2.0)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(3.0)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(4.0)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], "SA(5.0)": [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4, 4.5, 5.0], }

[output]
individual_curves = true
"""  # noqa


def test_get_iml():
    config = OpenquakeConfig()
    imts = ["PGA", "SA(0.5)"]
    imtls = list(range(10))
    config.set_iml(imts, imtls)
    assert (imts, imtls) == config.get_iml()


def test_get_iml_disagg():
    config = OpenquakeConfig()
    imt = "PGA"
    imtl = 0.1
    config.set_iml_disagg(imt, imtl)
    assert (imt, imtl) == config.get_iml_disagg()


def test_is_complete():
    config = OpenquakeConfig()
    assert not config.is_complete()

    imts = ["PGA", "SA(0.5)"]
    imtls = list(range(10))
    config.set_iml(imts, imtls)
    assert config.is_complete()

    config = OpenquakeConfig()
    imt = "PGA"
    imtl = 0.1
    config.set_iml_disagg(imt, imtl)
    assert config.is_complete()


def test_str():
    config = OpenquakeConfig()
    assert str(config)


def test_config_from_runzi(iml):
    config = (
        OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
        .set_description("hello world")
        .set_site_filepath("./sites.csv")
        .set_source_logic_tree_file("./hello.slt")
        .set_gsim_logic_tree_file("./gsim_model.xml")
        .set_uniform_site_params(750)
    )
    config.set_iml(iml['imts'], iml['imtls'])

    assert config.config['general']['description'] == "hello world"


def test_get(iml):
    config = (
        OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
        .set_description("hello world")
        .set_site_filepath("./sites.csv")
        .set_source_logic_tree_file("./hello.slt")
        .set_gsim_logic_tree_file("./gsim_model.xml")
        .set_uniform_site_params(750)
    )
    config.set_iml(iml['imts'], iml['imtls'])

    assert config.get_iml() == (iml['imts'], iml['imtls'])
    assert str(config.get_site_filepath()) == "sites.csv"
    assert config.get_uniform_site_params()[0] == 750
    assert config.set_parameter("foo", "bar", "foobar")
    assert config.get_parameter("foo", "bar") == "foobar"

    config.clear_iml()
    assert not config.get_iml()


def test_configuration_round_trip(iml):

    nc = (
        OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
        .set_site_filepath('./sites.csv')
        .set_parameter("general", "ps_grid_spacing", "20")
    )

    # levels1 = 'logscale(0.005, 4.00, 30)'

    nc.set_iml(iml['imts'], iml['imtls'])
    nc.set_uniform_site_params(250)
    nc.set_parameter("erf", "rupture_mesh_spacing", "42")

    out = io.StringIO()  # aother fake file
    nc.write(out)
    out.seek(0)
    nc2 = OpenquakeConfig.read_file(out)

    assert nc == nc2

    out.seek(0)
    for n, line in enumerate(out.readlines()):
        print(n, line, end="")

    print()
    print(str(nc))
    print(nc.config.get('general', 'random_seed'))
    # assert 0


def test_default_config():
    config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)  # the default config
    expected = configparser.ConfigParser()
    expected.read_dict(DEFAULT_HAZARD_CONFIG)
    assert config.config == expected


def test_set_maximum_distance():
    config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)  # the default config
    dists = {'Active Shallow Crust': 300.0, 'Volcanic': 300, 'Subduction Interface': 400, 'default': 400}
    config.set_maximum_distance(dists)
    assert config.config.get("calculation", "maximum_distance") == str(dists)


@pytest.fixture(scope='function')
def example_config():
    example = configparser.ConfigParser()
    example.read_file(io.StringIO(min_expected))
    config = OpenquakeConfig(example)
    yield config


def test_uniform_site_params():
    # default z values
    config = OpenquakeConfig()
    vs30_in = 100
    config.set_uniform_site_params(vs30_in)
    vs30, z1pt0, z2pt5 = config.get_uniform_site_params()
    assert vs30 == vs30_in
    assert z1pt0 == pytest.approx(round(oqhc.calculate_z1pt0(vs30_in), 0))
    assert z2pt5 == pytest.approx(round(oqhc.calculate_z2pt5(vs30_in), 1))

    # user specified z values
    config = OpenquakeConfig()
    vs30_in, z1pt0_in, z2pt5_in = (100, 490.0, 2.2)
    config.set_uniform_site_params(vs30_in, z1pt0_in, z2pt5_in)
    vs30, z1pt0, z2pt5 = config.get_uniform_site_params()
    assert vs30 == vs30_in
    assert z1pt0 == pytest.approx(round(z1pt0_in, 0))
    assert z2pt5 == pytest.approx(round(z2pt5_in, 1))


def test_unset_uniform_site_params():
    config = OpenquakeConfig()
    vs30_in = 100
    config.set_uniform_site_params(vs30_in)
    config.unset_uniform_site_params()
    vs30, z1pt0, z2pt5 = config.get_uniform_site_params()
    assert vs30 is None
    assert z1pt0 is None
    assert z2pt5 is None


def test_site_errors(locations):
    config = OpenquakeConfig()
    n_locs = len(locations)

    # vs30 values must be Iterable type
    with pytest.raises(TypeError):
        config.set_sites(locations, vs30=1)

    # vs30 must have same lenth as locations
    with pytest.raises(ValueError):
        config.set_sites(locations, vs30=[100] * (n_locs + 1))

    # cannot set site specific vs30 if uniform is specified
    config = OpenquakeConfig()
    config.set_uniform_site_params(100)

    assert config.set_sites(locations)

    vs30 = list(range(n_locs))
    with pytest.raises(KeyError):
        config.set_sites(locations, vs30=vs30)

    # cannot set uniform vs30 if site specific
    config = OpenquakeConfig()
    config.set_sites(locations, vs30=vs30)
    with pytest.raises(KeyError):
        config.set_uniform_site_params(100)


class TestConfigCompatability:
    def test_compatible_check_invariants(self, example_config):
        assert check_invariants(example_config.config)

    def test_compatible_check_invariant_missing_raises(self, example_config):
        example_config.config.remove_option('general', 'calculation_mode')
        with pytest.raises(ValueError):
            check_invariants(example_config.config)

    def test_compatible_check_invariant_bad_value_raises(self, example_config):
        example_config.config.set('general', 'calculation_mode', 'freddy flinstone')
        with pytest.raises(ValueError):
            check_invariants(example_config.config)

    def test_compatible_config(self, example_config):
        example_config.set_site_filepath('./sites.csv')
        compat = compatible_config(example_config.config)
        assert compat.get('site_params', 'sites_csv', fallback=None) is None
        assert compat.get('site_params', 'site_model_file', fallback=None) is None

    def test_compatible_hash_digest(self, example_config):
        example_config.set_site_filepath('./sites.csv')
        digest = compatible_hash_digest(example_config.config)
        print(digest)
        assert digest == "06f026df641e"

    def test_class_instance_hash_digest(self, example_config):
        assert example_config.compatible_hash_digest() == "06f026df641e"


@pytest.mark.skip("toml is not fully compatible due to use of quoted strings. Let's check if openquake supports this")
def test_toml_conversion():
    config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)  # the default config

    out = io.StringIO()
    config.write(out)
    out.seek(0)

    for n, line in enumerate(out.readlines()):
        print(n, line)

    tc = tomli.loads(out.getvalue())  # noqa

    assert tc['general']['random_seed'] == 25

    """
    assert 0
    io1 = io.StringIO()
    io2 = io.StringIO()
    self.config.write(io1)
    other.config.write(io2)


    # return tomli.loads(io1.getvalue()) == tomli.loads(io2.getvalue())
    # print('io1', len(io1.getvalue()), io1.getvalue()[:10], f'|{io1.getvalue()[-10:]}|')
    # print('io2', len(io2.getvalue()), io2.getvalue()[:10], f'|{io2.getvalue()[-10:]}|')
    print('io1', io1.getvalue().replace(' ', '.'))
    print('io2', io2.getvalue().replace(' ', '.'))

    d = difflib.Differ()
    result = list(d.compare(io1.getvalue().splitlines(keepends=True), io2.getvalue().splitlines(keepends=True)))
    pprint(result)

    return io2.getvalue() == io1.getvalue()
    """


def test_write_read_oq_config(tmp_path):
    config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)

    json_filepath = tmp_path / 'hazard_config.json'
    config.to_json(json_filepath)

    config_from_file = OpenquakeConfig().from_json(json_filepath)

    assert config_from_file.config.sections() == config.config.sections()
    for section in config.config.sections():
        assert config_from_file.config.options(section) == config.config.options(section)
        for option in config.config.options(section):
            assert config_from_file.config.get(section, option) == config.config.get(section, option)

    assert config_from_file.locations == config.locations
    assert config_from_file._site_parameters == config._site_parameters


def test_write_read_oq_config_site_params(tmp_path):
    config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
    locations = [CodedLocation(lat, lon, 0.001) for lat, lon in zip(range(10), range(10))]
    vs30s = [v / 10.0 for v in range(len(locations))]
    config.set_sites(locations, vs30=vs30s)

    json_filepath = tmp_path / 'hazard_config.json'
    config.to_json(json_filepath)

    config_from_file = OpenquakeConfig().from_json(json_filepath)

    assert config_from_file.config.sections() == config.config.sections()
    for section in config.config.sections():
        assert config_from_file.config.options(section) == config.config.options(section)
        for option in config.config.options(section):
            assert config_from_file.config.get(section, option) == config.config.get(section, option)

    assert config_from_file._site_parameters == config._site_parameters
    assert config_from_file.locations == config.locations



"""
A registry for logic_tree_branches (sources and GMMs) that are uniquely identified by a shake_256 hash.

Data is stored in the resources folder in CSV form. New branched may be added by developers to support
new source or logic tree components.

Examples:
    >>> from nzshm_model import branch_registry
    >>> registry = branch_registry.Registry()
    >>> entry = registry.source_registry.get_by_hash("af9ec2b004d7")
    ... BranchRegistryEntry(... )

Functions:
    identity_digest: get a standard hash_digest from an identity string
"""

import csv
import hashlib
import importlib.resources as resources
from dataclasses import asdict, dataclass
from typing import IO, Any, Optional, Union

HEADERS = ['hash_digest', 'identity', 'extra']
RESOURCES_DIR = resources.files('nzshm_model.resources')
GMM_REGISTRY_CSV = RESOURCES_DIR / 'gmm_branches.csv'
SOURCE_REGISTRY_CSV = RESOURCES_DIR / 'source_branches.csv'


def identity_digest(identity: str) -> str:
    """get a standard hash_digest from an identity string

    Arguments:
        identity: a unique identity string

    Returns:
        a shake_256 hash digest
    """
    return hashlib.shake_256(identity.encode()).hexdigest(6)


class Registry:
    """A container for the published registries.

    Attributes:
        gmm_registry: the model sources registry.
        source_registry: the model gmms registry.
    """

    _gmms: Optional['BranchRegistry'] = None
    _sources: Optional['BranchRegistry'] = None

    @property
    def gmm_registry(self) -> 'BranchRegistry':
        if not self._gmms:
            self._gmms = BranchRegistry().load(GMM_REGISTRY_CSV.open('r'))
        return self._gmms

    @property
    def source_registry(self) -> 'BranchRegistry':
        if not self._sources:
            self._sources = BranchRegistry().load(SOURCE_REGISTRY_CSV.open('r'))
        return self._sources


@dataclass
class BranchRegistryEntry:
    """A standard registry entry

    Attributes:
        identity: the unique identity string (see Branch.registry_identity)
        hash_digest: the shake_256 hexdigest of the identity.
        extra: more information about the entry.
    """

    identity: str
    hash_digest: Optional[str] = None
    extra: Optional[str] = None

    def __post_init__(self):
        if self.hash_digest:
            if not self.hash_digest == identity_digest(self.identity):
                raise ValueError(f'Incorrect hash_digest "{self.hash_digest}"" for "{self.identity}"')
        else:
            self.hash_digest = identity_digest(self.identity)


class BranchRegistry:
    """Storage Manager for BranchRegistryEntry objects"""

    def __init__(self):
        self._branches_by_hash = dict()
        self._branches_by_identity = dict()
        self._branches_by_extra = dict()

    def _load_row(self, row):
        entry = BranchRegistryEntry(**row)
        assert entry.hash_digest not in self._branches_by_hash
        assert entry.identity not in self._branches_by_identity

        self._branches_by_hash[entry.hash_digest] = entry
        self._branches_by_identity[entry.identity] = entry
        if entry.extra:
            self._branches_by_extra[entry.extra] = entry

    def load(self, registry_file: IO[Any]) -> 'BranchRegistry':
        """Load the entries contained in a CSV file.

        Arguments:
            registry_file: file-like object with expected CSV header file

        Returns:
            the populated BranchRegistry
        """
        registry_file.seek(0)
        reader = csv.DictReader(registry_file, fieldnames=HEADERS)
        headers = list(next(reader).values())
        assert HEADERS == headers
        for row in reader:
            self._load_row(row)
        return self

    def save(self, registry_file: IO[Any]) -> None:
        """Save the registry entries in CSV format.

        Arguments:
            registry_file: file-like object to write
        """
        csv_writer = csv.DictWriter(registry_file, fieldnames=HEADERS)
        csv_writer.writeheader()
        for row in self._branches_by_hash.values():
            csv_writer.writerow(asdict(row))

    def add(self, entry: BranchRegistryEntry) -> None:
        """add a new registry entry.

        Arguments:
            entry: a BranchRegistryEntry object.
        """
        if entry.hash_digest in self._branches_by_hash:
            assert entry.identity in self._branches_by_identity
        else:
            assert entry.identity not in self._branches_by_identity

        self._branches_by_hash[entry.hash_digest] = entry
        self._branches_by_identity[entry.identity] = entry

    def get_by_hash(self, hash_digest: str) -> BranchRegistryEntry:
        """Get a registry entry by hash_digest.

        Arguments:
            hash_digest: the hash digest string.
        """
        return self._branches_by_hash[hash_digest]

    def get_by_identity(self, identity: str) -> BranchRegistryEntry:
        """Get a registry entry by identity string.

        Arguments:
            identity: the identity string.
        """
        return self._branches_by_identity[identity]

    def get_by_extra(self, extra: str) -> Union[BranchRegistryEntry, None]:
        """Get a registry entry by the extra string.

        Notes:
         - this may return None.
         - only used a workaroound becuase some post NSHM hazard jobs used the extra value
           instead of the identity string in the HDF% ( e.g. `T3BlbnF1YWtlSGF6YXJkVGFzazo2OTMxODkz` )

        Arguments:
            extra: the extra string.
        """
        return self._branches_by_extra.get(extra)

    def __len__(self):
        return len(self._branches_by_hash.keys())



from dataclasses import dataclass, field
from typing import Any, Dict, Generator, List

from .branch_attribute import BranchAttributeSpec, BranchAttributeValue


@dataclass
class BranchSetSpec:
    short_name: str
    long_name: str
    branches: List['BranchAttributeValue'] = field(default_factory=list)


@dataclass
class BranchSetBase:
    short_name: str
    long_name: str
    branches: List[Any] = field(
        default_factory=list
    )  # we use 'Any' so subclassed version implementations will override typing

    def validate_weights(self) -> bool:
        weight = 0.0
        for b in self.branches:
            weight += b.weight
        return weight == 1.0

    def derive_spec(self) -> BranchSetSpec:
        fslt_spec = BranchSetSpec(short_name=self.short_name, long_name=self.long_name)

        options: Dict[str, set] = {}

        # extract unique values in to options
        for branch in self.branches:
            # iterate all the branches, yielding the unique  branch options
            # print('branch', branch)
            for value in branch.values:
                if value.name not in options.keys():
                    options[value.name] = set([])
                val = value.value
                if isinstance(val, list):  # make it hashable
                    val = tuple(val)
                bao = BranchAttributeValue(value.name, value.long_name, val)
                options[value.name].add(bao)

        # print(options)

        def option_values(options: Dict) -> Generator:
            # boil down the options values
            ret = {}
            for key, opt in options.items():
                if key not in ret:
                    ret[key] = BranchAttributeSpec(key, list(opt)[0].long_name)

                values = []
                for val in opt:
                    values.append(val.value)
                ret[key].value_options = sorted(values)

            for key, bas in ret.items():
                yield bas

        fslt_spec.branches = list(option_values(options))

        return fslt_spec



import csv
import warnings

import pytest

from nzshm_model.psha_adapter.openquake.hazard_config import OpenquakeConfig
from nzshm_model.psha_adapter.openquake.hazard_config_compat import DEFAULT_HAZARD_CONFIG
from nzshm_model.psha_adapter.openquake.simple_nrml import (
    OpenquakeConfigPshaAdapter,
    OpenquakeGMCMPshaAdapter,
    OpenquakeSourcePshaAdapter,
)


def test_source_adapter(tmp_path, current_model, source_map):
    cache_folder = tmp_path / 'cache'
    target_folder = tmp_path / 'target'
    source_adapter = current_model.source_logic_tree.psha_adapter(OpenquakeSourcePshaAdapter)
    source_filepath = source_adapter.write_config(cache_folder, target_folder, source_map)
    assert source_filepath.exists()


def test_gmcm_adapter(tmp_path, current_model):
    target_folder = tmp_path / 'target'
    gmcm_adapter = current_model.gmm_logic_tree.psha_adapter(OpenquakeGMCMPshaAdapter)
    gmcm_filepath = gmcm_adapter.write_config(target_folder)
    assert gmcm_filepath.exists()


def test_config_adapter(tmp_path, locations):
    target_folder = tmp_path / 'target'
    hazard_config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
    hazard_config.set_sites(locations)
    config_adapter = hazard_config.psha_adapter(OpenquakeConfigPshaAdapter)
    job_filepath = config_adapter.write_config(target_folder)
    assert (target_folder / 'sites.csv').exists()
    assert job_filepath.exists()


def test_config_sitefile(tmp_path, locations):
    site_file = tmp_path / 'sites1.csv'
    hazard_config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
    vs30s = [v / 10.0 for v in range(len(locations))]
    hazard_config.set_sites(locations, vs30=vs30s)
    config_adapter = hazard_config.psha_adapter(OpenquakeConfigPshaAdapter)
    config_adapter.write_site_file(site_file)
    assert site_file.exists()
    with site_file.open() as fin:
        reader = csv.reader(fin)
        header = next(reader)
        assert header == ['lon', 'lat', 'vs30']
        for loc, vs30, row in zip(locations, vs30s, reader):
            assert row[0] == str(loc.lon)
            assert row[1] == str(loc.lat)
            assert row[2] == str(vs30)


@pytest.mark.filterwarnings("default")
def test_write_config_warn(tmp_path, locations):
    target_folder = tmp_path / 'target'
    hazard_config = OpenquakeConfig(DEFAULT_HAZARD_CONFIG)
    hazard_config.set_sites(locations)
    config_adapter = hazard_config.psha_adapter(OpenquakeConfigPshaAdapter)
    with warnings.catch_warnings(record=True) as wngs:
        config_adapter.write_config(target_folder)
        assert len(wngs) > 0, len(wngs)
        assert "not complete" in str(wngs[-1].message)



from dataclasses import dataclass

from nzshm_model.psha_adapter.openquake.hazard_config import OpenquakeConfig

from .nshm_v1_0_0 import model_args_factory as model_args_nshm_v1_0_0
from .nshm_v1_0_4 import model_args_factory as model_args_nshm_v1_0_4

versions = {
    "NSHM_v1.0.0": model_args_nshm_v1_0_0,
    "NSHM_v1.0.4": model_args_nshm_v1_0_4,
}



import json
from pathlib import Path
from typing import Dict, Union

from nzshm_model.psha_adapter.hazard_config import HazardConfig
from nzshm_model.psha_adapter.openquake.hazard_config import OpenquakeConfig


class HazardConfigFactory:
    """
    Factory class used to get a HazardConfig type object without knowing the concrete type before runtime.

    Examples:
        >>> HazardConfigType = hazard_config_class_factory.get_hazard_config_class('openquake')
        >>> hazard_config = HazardConfigType()
        >>> HazardConfigType = hazard_config_class_factory.get_hazard_config_class_from_file(hazard_config.json)
        >>> hazard_config = HazardConfigType.from_json(hazard_config.json)
    """

    def __init__(self):
        self._config_types: Dict[str, type[HazardConfig]] = {}

    def register_config_class(self, type_name: str, config_type: type[HazardConfig]):
        self._config_types[type_name.casefold()] = config_type

    def get_hazard_config_class(self, type_name: str) -> type[HazardConfig]:
        """Get the concrete HazardConfig object."""
        config_type = self._config_types.get(type_name.casefold())
        if not config_type:
            raise ValueError(type_name)
        return config_type

    @staticmethod
    def detect_hazard_config_class_from_file(file_path: Union[str, Path]) -> str:
        """Get the concrete HazardConfig type from the json file representation of the object."""
        with Path(file_path).open() as config_file:
            data = json.load(config_file)
            return data["hazard_type"].casefold()

    def get_hazard_config_class_from_file(self, file_path: Union[str, Path]) -> type[HazardConfig]:
        """Get the concrete HazardConfig object by detecting the type from the json file
        representation of the object."""
        type_name = self.detect_hazard_config_class_from_file(file_path)
        return self.get_hazard_config_class(type_name)


hazard_config_class_factory = HazardConfigFactory()
hazard_config_class_factory.register_config_class('openquake', OpenquakeConfig)



from pathlib import Path

import pytest
from nzshm_common import CodedLocation


@pytest.fixture
def source_map(current_model):
    smap = {}
    for branch in current_model.source_logic_tree:
        for source in branch.sources:
            nrml_id = source.nrml_id
            smap[nrml_id] = [
                Path(f"path/to/{nrml_id}.xml"),
            ]
    return smap


@pytest.fixture
def locations():
    return [CodedLocation(lat, lon, 0.001) for lat, lon in zip(range(10), range(10))]


@pytest.fixture
def iml():
    _4_sites_levels = [
        0.01,
        0.02,
        0.04,
        0.06,
        0.08,
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
        0.8,
        0.9,
        1.0,
        1.2,
        1.4,
        1.6,
        1.8,
        2.0,
        2.2,
        2.4,
        2.6,
        2.8,
        3.0,
        3.5,
        4,
        4.5,
        5.0,
    ]
    _4_sites_measures = [
        'PGA',
        "SA(0.1)",
        "SA(0.2)",
        "SA(0.3)",
        "SA(0.4)",
        "SA(0.5)",
        "SA(0.7)",
        "SA(1.0)",
        "SA(1.5)",
        "SA(2.0)",
        "SA(3.0)",
        "SA(4.0)",
        "SA(5.0)",
    ]
    return {'imts': _4_sites_measures, 'imtls': _4_sites_levels}



import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Type, TypeVar, Union

from nzshm_model.psha_adapter import ConfigPshaAdapterInterface

if TYPE_CHECKING:
    from nzshm_common import CodedLocation

HazardConfigType = TypeVar('HazardConfigType', bound='HazardConfig')


class HazardConfig(ABC):

    hazard_type = "abstract"  # every child class should specify its type for use in loading from file

    def __init__(self, *args, **kwargs):
        pass

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def set_sites(self, locations: Sequence['CodedLocation'], **kwargs) -> 'HazardConfig':
        pass

    def to_json(self, file_path: Union[Path, str]) -> None:
        """serialized HazardConfig object to json file

        Args:
            file_path: path to file to be written
        """

        data = self.to_dict()
        with Path(file_path).open('w') as jsonfile:
            json.dump(data, jsonfile, indent=2)
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Convert HazardConfig object to dict

        Returns:
            dict representation of object
        """
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls: Type[HazardConfigType], data: Dict) -> HazardConfigType:
        """
        create a HazardConfig object from a dictionary

        Args:
            data: the dictionary representation of the HazardConfig object

        Returns:
            a HazardConfig object
        """
        pass

    @classmethod
    @abstractmethod
    def from_json(cls: Type[HazardConfigType], file_path: Union[Path, str]) -> HazardConfigType:
        pass

    def psha_adapter(
        self, provider: Type[ConfigPshaAdapterInterface], **kwargs: Optional[Dict]
    ) -> "ConfigPshaAdapterInterface":
        """get a PSHA adapter for this instance.

        Arguments:
            provider: the adapter class
            **kwargs: additional arguments required by the provider class

        Returns:
            a PSHA Adapter instance
        """
        return provider(target=self)



import os

import botocore.exceptions

try:
    from nshm_toshi_client.toshi_file import ToshiFile

    from nzshm_model.logic_tree.source_logic_tree.toshi_api import get_secret
except (ModuleNotFoundError, ImportError):
    print("WARNING: optional `toshi` dependencies are not installed.")
    raise

# Get API key from AWS secrets manager
API_URL = os.getenv('NZSHM22_TOSHI_API_URL', "http://127.0.0.1:5000/graphql")
try:
    if 'TEST' in API_URL.upper():
        API_KEY = get_secret("NZSHM22_TOSHI_API_SECRET_TEST", "us-east-1").get("NZSHM22_TOSHI_API_KEY_TEST")
    elif 'PROD' in API_URL.upper():
        API_KEY = get_secret("NZSHM22_TOSHI_API_SECRET_PROD", "us-east-1").get("NZSHM22_TOSHI_API_KEY_PROD")
    else:
        API_KEY = os.getenv('NZSHM22_TOSHI_API_KEY', "")
except (AttributeError, botocore.exceptions.EndpointConnectionError) as err:
    print(f"unable to get secret from secretmanager: {err}")
    API_KEY = os.getenv('NZSHM22_TOSHI_API_KEY', "")

# S3_URL = None
# DEPLOYMENT_STAGE = os.getenv('DEPLOYMENT_STAGE', 'LOCAL').upper()
# REGION = os.getenv('REGION', 'ap-southeast-2')  # SYDNEY


class SourceSolution(ToshiFile):
    def get_source(self, fid):

        qry = '''
        query file ($id:ID!) {
            node(id: $id) {
                __typename
                ... on FileInterface {
                  file_name
                  file_size
                  meta {k v}
                }
            }
        }
        '''
        # print(qry)
        input_variables = dict(id=fid)
        executed = self.run_query(qry, input_variables)
        return executed['node']



"""
Defines ground motion characterisation model logic tree structures used in NSHM.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, List

from nzshm_model.logic_tree.logic_tree_base import Branch, BranchSet, FilteredBranch, LogicTree


@dataclass
class GMCMBranch(Branch):
    """
    A branch of the GMCM logic tree

    Identifies the ground motion model (gsim) and any arguments


    Attributes:
        gsim_name: the name of the ground motion model
        gsim_args: a dict of argument names and values
    """

    gsim_name: str = ""
    gsim_args: Dict[str, Any] = field(default_factory=dict)
    tectonic_region_type: str = ""  # need a default becasue base class has a memeber with a default

    def filtered_branch(self, logic_tree, branch_set) -> 'GMCMFilteredBranch':
        """get a filtered branch containing reference to parent instances.

        Arguments:
            logic_tree: the gmcm logic tree containing the branch
            branch_set: the branch set containing the branch

        Returns:
            a BMCMFilteredBranch instance
        """
        return GMCMFilteredBranch(logic_tree=logic_tree, branch_set=branch_set, **self.__dict__)

    @property
    def registry_identity(self):
        arg_vals = []
        for k, v in self.gsim_args.items():
            arg_vals.append(f"{k}={v}")
        return f"{self.gsim_name}({', '.join(arg_vals)})"


# TODO: protect from users changing TRT
@dataclass
class GMCMBranchSet(BranchSet[GMCMBranch]):
    """A list of GMCM branches that apply to a particular tectonic region.

    Attributes:
        tectonic_region_type: the tectonic region type TRT that the branch set applies to.
        branches: list of branches.
    """

    branches: List[GMCMBranch] = field(default_factory=list)

    def __post_init__(self):
        trts = {branch.tectonic_region_type for branch in self.branches}
        if len(trts) > 1:
            raise ValueError("all tectonic_region_types in a branch set must be the same")

    @property
    def tectonic_region_type(self) -> str:
        return self.branches[0].tectonic_region_type if self.branches else ""


@dataclass
class GMCMLogicTree(LogicTree['GMCMFilteredBranch']):
    """A dataclass representing the ground motion characterisation model logic tree.

    Attributes:
        branch_sets: list of GMCM branch sets that comprise the logic tree.
    """

    # should we enforce that there is only one branch_set per TRT?
    branch_sets: List[GMCMBranchSet] = field(default_factory=list)

    def __post_init__(self):
        self._fix_args()

    def _fix_args(self) -> 'GMCMLogicTree':
        """Replace string representations of numeric arguments with floats"""

        def is_number(value):
            return value.replace("-", "").replace(".", "").replace("e", "").isnumeric()

        for branch_set in self.branch_sets:
            for branch in branch_set.branches:
                for k, v in branch.gsim_args.items():
                    if (isinstance(v, str)) and (is_number(v)):
                        branch.gsim_args[k] = float(v)

        return self


@dataclass
class GMCMFilteredBranch(FilteredBranch, GMCMBranch):
    logic_tree: 'GMCMLogicTree' = field(default_factory=GMCMLogicTree)
    branch_set: 'GMCMBranchSet' = field(default_factory=GMCMBranchSet)



from nzshm_model.psha_adapter.openquake import OpenquakeModelPshaAdapter

imtls = list(range(10))
imts = ["PGA", "SA(1.0)"]


def test_write_config(tmp_path, current_model, source_map):
    cache_folder = tmp_path / 'cache'
    target_folder = tmp_path / 'target'

    current_model.hazard_config.set_site_filepath('sites.csv')
    current_model.hazard_config.set_iml(imts, imtls)

    current_model.psha_adapter(OpenquakeModelPshaAdapter).write_config(cache_folder, target_folder, source_map)
    assert (target_folder / 'job.ini').exists()
    assert (target_folder / 'gsim_model.xml').exists()
    assert (target_folder / 'sources' / 'sources.xml').exists()



#! branch_attribute.py

"""
Branch structures used in NSHM source logic trees
"""

from dataclasses import dataclass, field
from typing import Any, List


@dataclass
class BranchAttributeSpec:
    name: str
    long_name: str
    value_options: List[Any] = field(default_factory=list)


@dataclass(frozen=True)
class BranchAttributeValue:
    name: str
    long_name: str
    value: Any = None

    @staticmethod
    def from_branch_attribute(ba: BranchAttributeSpec, value):
        return BranchAttributeValue(ba.name, ba.long_name, value)

    @staticmethod
    def all_from_branch_attribute(ba: BranchAttributeSpec):
        for opt in ba.value_options:
            yield BranchAttributeValue(ba.name, ba.long_name, opt)

    def __repr__(self):
        return f"{self.name}{self.value}"



import importlib.resources as resources
from pathlib import Path

import pytest

from nzshm_model.psha_adapter.hazard_config_factory import hazard_config_class_factory
from nzshm_model.psha_adapter.openquake.hazard_config import OpenquakeConfig

fixtures_dir = resources.files('tests.psha_adapter.fixtures')


def test_types():

    hazard_config_class_factory.get_hazard_config_class('openquake')
    hazard_config_class_factory.get_hazard_config_class('OpEnQuAkE')


def test_wrong_type():

    with pytest.raises(ValueError):
        hazard_config_class_factory.get_hazard_config_class('microsoft word')


def test_class_name_from_file():

    config_filepath = Path(fixtures_dir / 'hazard_config.json')
    config_type = hazard_config_class_factory.detect_hazard_config_class_from_file(config_filepath)
    assert config_type == 'openquake'


def test_class_from_file():

    config_filepath = Path(fixtures_dir / 'hazard_config.json')
    config_class = hazard_config_class_factory.get_hazard_config_class_from_file(config_filepath)
    assert config_class == OpenquakeConfig



# from .openquake import NrmlDocument, OpenquakeSimplePshaAdapter
from .psha_adapter_interface import (
    ConfigPshaAdapterInterface,
    GMCMPshaAdapterInterface,
    ModelPshaAdapterInterface,
    PshaAdapterInterface,
    SourcePshaAdapterInterface,
)



"""
This package provides a psha_adapter implementation for the openquake PSHA.

Classes:
    OpenquakeConfig: a concrete psha_adapter for openquake hazard configuration

Attributes:
    DEFAULT_HAZARD_CONFIG: the base hazard configuration for NZHSM

"""

from .hazard_config import OpenquakeConfig
from .hazard_config_compat import DEFAULT_HAZARD_CONFIG
from .logic_tree import NrmlDocument
from .simple_nrml import (
    OpenquakeConfigPshaAdapter,
    OpenquakeGMCMPshaAdapter,
    OpenquakeModelPshaAdapter,
    OpenquakeSourcePshaAdapter,
    gmcm_branch_from_element_text,
)
